--"logical.vhd"; SIMSYNCH logical utility functions for VHDL.
-- Copyright (C) 2001, 2002 Aubrey Jaffer
--
-- Permission to copy this VHDL library, to modify it, to redistribute
-- it, to distribute modified versions, and to use it for any purpose is
-- granted, subject to the following restrictions and understandings.
--
-- 1.  Any copy made of this library must include this copyright notice
-- in full.
--
-- 2.  I have made no warranty or representation that the operation of
-- this library will be error-free, and I am under no obligation to
-- provide any services, by way of maintenance, update, or otherwise.
--
-- 3.  In conjunction with products arising from the use of this
-- material, there shall be no use of my name in any advertising,
-- promotional, or sales literature without prior written consent in
-- each case.

LIBRARY ieee;

USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_arith.ALL;
USE ieee.std_logic_unsigned.ALL;
--
-- FIF(test, consequent, alternate) is Functional-IF,
--  useful in contexts where VHDL disallows IF statements.
--
--  If TEST is True or '1', then return CONSEQUENT;
--  otherwise return ALTERNATE.
--
-- LOGBIT_P(index, operand)
--  returns True or '1' if OPERAND(INDEX) is '1';
--  otherwise returns False or '0'.
--
-- BIT_FIELD(operand, istart, iend)
--  returns OPERAND(IEND downto ISTART).
--
-- RSHIFT(operand, icnt)
--  shifts (unsigned) OPERAND ICNT bits to the right;
--  The low-order bits are truncated.
-- LSHIFT(operand, icnt)
--  shifts OPERAND ICNT bits to the left;
--  The low-order bits are padded with '0'.
--
PACKAGE logical IS

  -- purpose: extract bit IND of word INT
  FUNCTION logbit_P (idx : natural; btv : std_logic_vector)
    RETURN boolean;
  FUNCTION logbit_P (idx : std_logic_vector; btv : std_logic_vector)
    RETURN boolean;
  FUNCTION logbit_P (idx : natural; int : integer)
    RETURN boolean;
  FUNCTION logbit_P (idx : std_logic_vector; int : integer)
    RETURN boolean;

  -- Functional IF
  FUNCTION fif(test : boolean; consequent, alternate : boolean)
    RETURN boolean;
  FUNCTION fif(test : std_logic; consequent, alternate : boolean)
    RETURN boolean;

  FUNCTION fif(test : boolean; consequent, alternate : std_logic)
    RETURN boolean;
  FUNCTION fif(test : std_logic; consequent, alternate : std_logic)
    RETURN boolean;

  FUNCTION fif(test : boolean; consequent : std_logic; alternate : boolean)
    RETURN boolean;
  FUNCTION fif(test : std_logic; consequent : boolean; alternate : std_logic)
    RETURN boolean;

  FUNCTION fif(test : boolean; consequent, alternate : std_logic_vector)
    RETURN std_logic_vector;
  FUNCTION fif(test : std_logic; consequent, alternate : std_logic_vector)
    RETURN std_logic_vector;
  FUNCTION fif(test : boolean; consequent, alternate : integer)
    RETURN integer;
  FUNCTION fif(test : std_logic; consequent, alternate : integer)
    RETURN integer;

  FUNCTION fif(test : boolean; consequent : integer; alternate : std_logic_vector)
    RETURN std_logic_vector;
  FUNCTION fif(test : std_logic; consequent : integer; alternate : std_logic_vector)
    RETURN std_logic_vector;
  FUNCTION fif(test : boolean; consequent : std_logic_vector; alternate : integer)
    RETURN std_logic_vector;
  FUNCTION fif(test : std_logic; consequent : std_logic_vector; alternate : integer)
    RETURN std_logic_vector;

  -- purpose: extract bit field ISTART to IEND from btv.
  FUNCTION bit_field (n : std_logic_vector; istart, iend: natural)
    RETURN std_logic_vector;
  FUNCTION bit_field (n : integer; istart, iend: natural)
    RETURN integer;
  FUNCTION bit_field (n : std_logic_vector; istart, iend: std_logic_vector)
    RETURN std_logic_vector;
  FUNCTION bit_field (n : integer; istart, iend: std_logic_vector)
    RETURN integer;

  -- purpose: right-shift.
  FUNCTION rShift (btv : std_logic_vector; icnt : natural)
    RETURN std_logic_vector;
  FUNCTION rShift (btv : integer; icnt : natural)
    RETURN integer;

  -- purpose: left-shift.
  FUNCTION lShift (btv : std_logic_vector; icnt : natural)
    RETURN std_logic_vector;
  FUNCTION lShift (btv : integer; icnt : natural)
    RETURN integer;

  FUNCTION To_StdLogicVector (slv: std_logic_vector) RETURN std_logic_vector;

  FUNCTION To_StdLogic(bt : boolean)
    RETURN std_logic;
  FUNCTION To_StdLogic(bt : std_logic)
    RETURN std_logic;

  -------------------------------------------------------------------
  -- integer overloaded logical operators
  -------------------------------------------------------------------
  FUNCTION "and"  (j, k : integer) RETURN integer;
  FUNCTION "and"  (j : integer; k : std_logic_vector) RETURN std_logic_vector;
  FUNCTION "and"  (j : std_logic_vector; k : integer) RETURN std_logic_vector;

  FUNCTION "nand" (j, k : integer) RETURN integer;
  FUNCTION "nand" (j : integer; k : std_logic_vector) RETURN std_logic_vector;
  FUNCTION "nand" (j : std_logic_vector; k : integer) RETURN std_logic_vector;

  FUNCTION "or"   (j, k : integer) RETURN integer;
  FUNCTION "or"   (j : integer; k : std_logic_vector) RETURN std_logic_vector;
  FUNCTION "or"   (j : std_logic_vector; k : integer) RETURN std_logic_vector;

  FUNCTION "nor"  (j, k : integer) RETURN integer;
  FUNCTION "nor"  (j : integer; k : std_logic_vector) RETURN std_logic_vector;
  FUNCTION "nor"  (j : std_logic_vector; k : integer) RETURN std_logic_vector;

  FUNCTION "xor"  (j, k : integer) RETURN integer;
  FUNCTION "xor"  (j : integer; k : std_logic_vector) RETURN std_logic_vector;
  FUNCTION "xor"  (j : std_logic_vector; k : integer) RETURN std_logic_vector;

  FUNCTION "not"  (j : integer) RETURN integer;

  FUNCTION Conv_Std_Logic_Vector(arg: STD_LOGIC_VECTOR; size: INTEGER)
    RETURN STD_LOGIC_VECTOR;

-- synopsys translate_off
  -------------------------------------------------------------------
  -- Remove mixed std_logic_vector/std_ulogic overloaded arithmetic operators
  -------------------------------------------------------------------
  FUNCTION "+"(L: STD_LOGIC_VECTOR; R: STD_ULOGIC) RETURN STD_LOGIC_VECTOR;
  FUNCTION "+"(L: STD_ULOGIC; R: STD_LOGIC_VECTOR) RETURN STD_LOGIC_VECTOR;
  FUNCTION "-"(L: STD_LOGIC_VECTOR; R: STD_ULOGIC) RETURN STD_LOGIC_VECTOR;
  FUNCTION "-"(L: STD_ULOGIC; R: STD_LOGIC_VECTOR) RETURN STD_LOGIC_VECTOR;
  FUNCTION Conv_Std_Logic_Vector(ARG: STD_ULOGIC; SIZE: INTEGER)
    RETURN STD_LOGIC_VECTOR;
  FUNCTION Conv_Unsigned(arg: STD_LOGIC_VECTOR; size: INTEGER)
    RETURN UNSIGNED;

  -------------------------------------------------------------------
  -- Check that each bit of vector is '0' or '1'.
  -------------------------------------------------------------------
  PROCEDURE Must_Be_Valid(slv : STD_LOGIC_VECTOR; CONSTANT what : STRING);
  PROCEDURE Must_Be_Valid(slv : STD_LOGIC_VECTOR;
                          CONSTANT what : STRING;
                          sevlev : SEVERITY_LEVEL);
-- synopsys translate_on

END logical;

PACKAGE BODY logical IS
-- synopsys translate_off
  -------------------------------------------------------------------
  -- Check that each bit of vector is '0' or '1'.
  -------------------------------------------------------------------
  PROCEDURE Must_Be_Valid(slv : STD_LOGIC_VECTOR; CONSTANT what : STRING) IS
  BEGIN
    Must_Be_Valid(slv, what, ERROR);
  END;

  PROCEDURE Must_Be_Valid(SLV: STD_LOGIC_VECTOR;
                          CONSTANT what : STRING;
                          sevlev : SEVERITY_LEVEL) IS
    VARIABLE tmp_char : character;
    VARIABLE tmp_str : STRING(1 TO SLV'length);
  BEGIN
    IF (10 ns > NOW) THEN RETURN; END IF;
    FOR idx IN SLV'RANGE LOOP
      IF NOT (SLV(idx) = '0' OR SLV(idx) = '1') THEN
        FOR sdx IN SLV'low TO slv'high LOOP
          CASE SLV(sdx) IS
            WHEN 'U'    => tmp_char := 'U';  -- Uninitialized
            WHEN 'X'    => tmp_char := 'X';  -- Forcing  Unknown
            WHEN '0'    => tmp_char := '0';  -- Forcing  0
            WHEN '1'    => tmp_char := '1';  -- Forcing  1
            WHEN 'Z'    => tmp_char := 'Z';  -- High Impedance
            WHEN 'W'    => tmp_char := 'W';  -- Weak     Unknown
            WHEN 'L'    => tmp_char := 'L';  -- Weak     0
            WHEN 'H'    => tmp_char := 'H';  -- Weak     1
            WHEN '-'    => tmp_char := '-';  -- Don't care
            WHEN OTHERS => tmp_char := '?';
          END CASE;
          tmp_str := tmp_char & tmp_str(1 TO tmp_str'length-1);

        END LOOP;
        ASSERT FALSE REPORT what & " = " & tmp_str SEVERITY sevlev;
        RETURN;
      END IF;
    END LOOP;
  END;
-- synopsys translate_on

  FUNCTION Min(l, r: INTEGER) RETURN INTEGER IS
  BEGIN
    IF l < r THEN
      RETURN l;
    ELSE
      RETURN r;
    END IF;
  END;

  FUNCTION Conv_Unsigned(arg: STD_LOGIC_VECTOR; size: INTEGER)
    RETURN UNSIGNED IS
    CONSTANT msb: INTEGER := Min(arg'LENGTH, size) - 1;
    SUBTYPE rtype IS UNSIGNED (size-1 DOWNTO 0);
    VARIABLE new_bounds: STD_LOGIC_VECTOR (arg'LENGTH-1 DOWNTO 0);
    VARIABLE result: rtype;
  BEGIN
    new_bounds := arg;
    result := rtype'(OTHERS => '0');
    FOR i IN msb DOWNTO 0 LOOP
      result(i) := new_bounds(i);
    END LOOP;
    RETURN result;
  END;

  FUNCTION Conv_Std_Logic_Vector(arg: STD_LOGIC_VECTOR; size: INTEGER)
    RETURN STD_LOGIC_VECTOR IS
  BEGIN
    RETURN Ext(arg, size);
  END;

-- synopsys translate_off
  -------------------------------------------------------------------
  -- Remove mixed std_logic_vector/std_ulogic overloaded arithmetic operators
  -------------------------------------------------------------------
  FUNCTION "+"(L: STD_LOGIC_VECTOR; R: STD_ULOGIC) RETURN STD_LOGIC_VECTOR IS
  BEGIN
    ASSERT false REPORT "Mixing logic and vector" SEVERITY FAILURE;
    RETURN L;
  END "+";
  FUNCTION "+"(L: STD_ULOGIC; R: STD_LOGIC_VECTOR) RETURN STD_LOGIC_VECTOR IS
  BEGIN
    ASSERT false REPORT "Mixing logic and vector" SEVERITY FAILURE;
    RETURN R;
  END "+";
  FUNCTION "-"(L: STD_LOGIC_VECTOR; R: STD_ULOGIC) RETURN STD_LOGIC_VECTOR IS
  BEGIN
    ASSERT false REPORT "Mixing logic and vector" SEVERITY FAILURE;
    RETURN L;
  END "-";
  FUNCTION "-"(L: STD_ULOGIC; R: STD_LOGIC_VECTOR) RETURN STD_LOGIC_VECTOR IS
  BEGIN
    ASSERT false REPORT "Mixing logic and vector" SEVERITY FAILURE;
    RETURN R;
  END "-";
  FUNCTION Conv_Std_Logic_Vector(ARG: STD_ULOGIC; SIZE: INTEGER) RETURN STD_LOGIC_VECTOR IS
    SUBTYPE rtype IS STD_LOGIC_VECTOR (SIZE-1 DOWNTO 0);
    VARIABLE result: rtype;
  BEGIN
    ASSERT false REPORT "Mixing logic and vector" SEVERITY FAILURE;
    result := rtype'(OTHERS => '0');
    RETURN result;
  END Conv_Std_Logic_Vector;
-- synopsys translate_on

  FUNCTION To_StdLogicVector (slv : std_logic_vector) RETURN std_logic_vector IS
  BEGIN
    RETURN slv;
  END;

  -------------------------------------------------------------------
  -- integer overloaded logical operators
  -------------------------------------------------------------------

  FUNCTION "and"  (j,k : integer) RETURN integer IS
  BEGIN
    if j = k then
      RETURN k;
    elsif (0 = j) or (0 = k) then
      RETURN 0;
    else
      RETURN 2 * (j/2 and k/2) + (j mod 2)*(k mod 2);
    end if;
  END "and";
  FUNCTION "and"  (j : integer; k : std_logic_vector) RETURN std_logic_vector IS
    VARIABLE btv : std_logic_vector(k'LENGTH-1 DOWNTO 0);
  BEGIN
    btv := Conv_Std_Logic_Vector(j, k'LENGTH);
    RETURN btv AND k;
  END "and";
  FUNCTION "and"  (j : std_logic_vector; k : integer) RETURN std_logic_vector IS
    VARIABLE btv : std_logic_vector(j'LENGTH-1 DOWNTO 0);
  begin
    btv := conv_std_logic_vector(k, j'LENGTH);
    RETURN j AND btv;
  END "and";

  FUNCTION "nand" (j, k : integer) RETURN integer IS
  BEGIN
    RETURN not (j and k);
  END "nand";
  FUNCTION "nand" (j : integer; k : std_logic_vector) RETURN std_logic_vector IS
    VARIABLE btv : std_logic_vector(k'LENGTH-1 DOWNTO 0);
  BEGIN
    btv := Conv_Std_Logic_Vector(j, k'LENGTH);
    RETURN btv NAND k;
  END "nand";
  FUNCTION "nand" (j : std_logic_vector; k : integer) RETURN std_logic_vector IS
    VARIABLE btv : std_logic_vector(j'LENGTH-1 DOWNTO 0);
  begin
    btv := conv_std_logic_vector(k, j'LENGTH);
    RETURN j NAND btv;
  END "nand";

  FUNCTION "or"  (j,k : integer) RETURN integer IS
  BEGIN
    if j = k then
      RETURN k;
    elsif (0 = j) or (0 = k) then
      RETURN 0;
    else
      RETURN 2 * (j/2 or k/2) + 1 - (j mod 2)*(k mod 2);
    end if;
  END "or";
  FUNCTION "or" (j : integer; k : std_logic_vector) RETURN std_logic_vector IS
    VARIABLE btv : std_logic_vector(k'LENGTH-1 DOWNTO 0);
  BEGIN
    btv := Conv_Std_Logic_Vector(j, k'LENGTH);
    RETURN btv OR k;
  END "or";
  FUNCTION "or" (j : std_logic_vector; k : integer) RETURN std_logic_vector IS
    VARIABLE btv : std_logic_vector(j'LENGTH-1 DOWNTO 0);
  BEGIN
    btv := Conv_Std_Logic_Vector(k, j'LENGTH);
    RETURN j OR btv;
  END "or";

  FUNCTION "nor" (j, k : integer) RETURN integer IS
  BEGIN
    RETURN not (j or k);
  END "nor";
  FUNCTION "nor" (j : integer; k : std_logic_vector) RETURN std_logic_vector IS
    VARIABLE btv : std_logic_vector(k'LENGTH-1 DOWNTO 0);
  BEGIN
    btv := Conv_Std_Logic_Vector(j, k'LENGTH);
    RETURN btv NOR k;
  END "nor";
  FUNCTION "nor" (j : std_logic_vector; k : integer) RETURN std_logic_vector IS
    VARIABLE btv : std_logic_vector(j'LENGTH-1 DOWNTO 0);
  BEGIN
    btv := Conv_Std_Logic_Vector(k, j'LENGTH);
    RETURN j NOR btv;
  END "nor";

  FUNCTION "xor"  (j,k : integer) RETURN integer IS
  BEGIN
    if j = k then
      RETURN 0;
    elsif 0 = j then
      RETURN k;
    elsif 0 = k then
      RETURN j;
    else
      RETURN 2 * (j/2 xor k/2) + (j mod 2)*(k mod 2);
    end if;
  END "xor";
  FUNCTION "xor" (j : integer; k : std_logic_vector) RETURN std_logic_vector IS
    VARIABLE btv : std_logic_vector(k'LENGTH-1 DOWNTO 0);
  BEGIN
    btv := Conv_Std_Logic_Vector(j, k'LENGTH);
    RETURN btv XOR k;
  END "xor";
  FUNCTION "xor" (j : std_logic_vector; k : integer) RETURN std_logic_vector IS
    VARIABLE btv : std_logic_vector(j'LENGTH-1 DOWNTO 0);
  BEGIN
    btv := Conv_Std_Logic_Vector(k, j'LENGTH);
    RETURN j XOR btv;
  END "xor";

  FUNCTION "not"  (j : integer) RETURN integer IS
  BEGIN
    RETURN (-1 - j);
  END "not";

  FUNCTION rShift (btv : integer; icnt : natural)
    RETURN integer IS
  BEGIN                                 -- rShift
    RETURN btv / (2 ** icnt);
  END rShift;
  FUNCTION rShift (btv : std_logic_vector; icnt : natural)
    RETURN std_logic_vector IS
    CONSTANT btv_msb: integer := btv'LENGTH - 1;
  BEGIN                                 -- rShift
    RETURN btv(btv_msb DOWNTO icnt);
  END rShift;

  FUNCTION lshift (btv : integer; icnt : natural)
    RETURN integer IS
  BEGIN                                 -- lshift
    RETURN btv * (2 ** icnt);
  END lshift;
  FUNCTION lshift (btv : std_logic_vector; icnt : natural)
    RETURN std_logic_vector IS
    CONSTANT btv_msb: integer := btv'LENGTH - 1 + icnt;
    VARIABLE bsv: std_logic_vector (btv_msb DOWNTO 0);
  BEGIN                                 -- lshift
    bsv := (others => '0');
    bsv(btv_msb DOWNTO icnt) := btv;
    RETURN bsv;
  END lshift;

  -- purpose: extract bit field ISTART to IEND from btv.
  FUNCTION bit_field (n : std_logic_vector; istart, iend: natural)
    RETURN std_logic_vector IS
  BEGIN
    RETURN n(iend DOWNTO istart);
  END bit_field;
  FUNCTION bit_field (n : integer; istart, iend: natural)
    RETURN integer IS
  BEGIN
    RETURN rshift(n, istart) AND (lshift(1, iend - istart) - 1);
  END bit_field;
  FUNCTION bit_field (n : std_logic_vector; istart, iend: std_logic_vector)
    RETURN std_logic_vector IS
  BEGIN
    RETURN n(Conv_Integer(iend) DOWNTO Conv_Integer(istart));
  END bit_field;
  FUNCTION bit_field (n : integer; istart, iend: std_logic_vector)
    RETURN integer IS
  BEGIN
    RETURN rshift(n, Conv_Integer(istart))
      AND (lshift(1, Conv_Integer(iend) - Conv_Integer(istart)) - 1);
  END bit_field;

  -- purpose: convert boolean or std_logic to std_logic.
  FUNCTION To_StdLogic(bt : boolean)
    RETURN std_logic IS
  BEGIN
    IF bt THEN
      RETURN '1';
    ELSE
      RETURN '0';
    END IF;
  END To_StdLogic;
  FUNCTION To_StdLogic(bt : std_logic)
    RETURN std_logic IS
  BEGIN
    RETURN bt;
  END To_StdLogic;

  -- purpose: extract bit IDX of word BTV
  FUNCTION logbit_P (idx : natural; btv : std_logic_vector)
    RETURN boolean IS
  BEGIN  -- logbit_P
    RETURN '1' = btv(idx);
  END logbit_P;
  FUNCTION logbit_P (idx : std_logic_vector; btv : std_logic_vector)
    RETURN boolean IS
  BEGIN  -- logbit_P
    RETURN '1' = btv(Conv_Integer(idx));
  END logbit_P;
  FUNCTION logbit_P (idx : natural; int : integer)
    RETURN boolean IS
    VARIABLE btv : std_logic_vector(31 DOWNTO 0);
  BEGIN  -- logbit_P
    btv := Conv_Std_Logic_Vector(int, 32);
    RETURN '1' = btv(idx);
  END logbit_P;
  FUNCTION logbit_P (idx : std_logic_vector; int : integer)
    RETURN boolean IS
    VARIABLE btv : std_logic_vector(31 DOWNTO 0);
  BEGIN  -- logbit_P
    btv := Conv_Std_Logic_Vector(int, 32);
    RETURN '1' = btv(Conv_Integer(idx));
  END logbit_P;

  FUNCTION fif(test : boolean; consequent, alternate : boolean)
    RETURN boolean IS
  BEGIN
    IF test THEN
      RETURN consequent;
    ELSE
      RETURN alternate;
    END IF;
  END fif;
  FUNCTION fif(test : std_logic; consequent, alternate : boolean)
    RETURN boolean IS
  BEGIN
    IF '1' = test THEN
      RETURN consequent;
    ELSE
      RETURN alternate;
    END IF;
  END fif;

  FUNCTION fif(test : boolean; consequent, alternate : std_logic)
    RETURN boolean IS
  BEGIN
    IF test THEN
      RETURN '1' = consequent;
    ELSE
      RETURN '1' = alternate;
    END IF;
  END fif;
  FUNCTION fif(test : std_logic; consequent, alternate : std_logic)
    RETURN boolean IS
  BEGIN
    IF '1' = test THEN
      RETURN '1' = consequent;
    ELSE
      RETURN '1' = alternate;
    END IF;
  END fif;

  FUNCTION fif(test : boolean; consequent : std_logic; alternate : boolean)
    RETURN boolean IS
  BEGIN
    IF test THEN
      RETURN '1' = consequent;
    ELSE
      RETURN alternate;
    END IF;
  END fif;
  FUNCTION fif(test : std_logic; consequent : boolean; alternate : std_logic)
    RETURN boolean IS
  BEGIN
    IF '1' = test THEN
      RETURN consequent;
    ELSE
      RETURN '1' = alternate;
    END IF;
  END fif;

  FUNCTION fif(test : boolean; consequent, alternate : std_logic_vector)
    RETURN std_logic_vector IS
  BEGIN
    IF test THEN
      RETURN consequent;
    ELSE
      RETURN alternate;
    END IF;
  END fif;
  FUNCTION fif(test : std_logic; consequent, alternate : std_logic_vector)
    RETURN std_logic_vector IS
  BEGIN
    IF '1' = test THEN
      RETURN consequent;
    ELSE
      RETURN alternate;
    END IF;
  END fif;
  FUNCTION fif(test : boolean; consequent, alternate : integer)
    RETURN integer IS
  BEGIN
    IF test THEN
      RETURN consequent;
    ELSE
      RETURN alternate;
    END IF;
  END fif;
  FUNCTION fif(test : std_logic; consequent, alternate : integer)
    RETURN integer IS
  BEGIN
    IF '1' = test THEN
      RETURN consequent;
    ELSE
      RETURN alternate;
    END IF;
  END fif;

  FUNCTION fif(test : boolean; consequent : integer; alternate : std_logic_vector)
    RETURN std_logic_vector IS
  BEGIN
    IF test THEN
      RETURN Conv_Std_Logic_Vector(consequent, alternate'LENGTH);
    ELSE
      RETURN alternate;
    END IF;
  END fif;
  FUNCTION fif(test : std_logic; consequent : integer; alternate : std_logic_vector)
    RETURN std_logic_vector IS
  BEGIN
    IF '1' = test THEN
      RETURN Conv_Std_Logic_Vector(consequent, alternate'LENGTH);
    ELSE
      RETURN alternate;
    END IF;
  END fif;
  FUNCTION fif(test : boolean; consequent : std_logic_vector; alternate : integer)
    RETURN std_logic_vector IS
  BEGIN
    IF test THEN
      RETURN consequent;
    ELSE
      RETURN Conv_Std_Logic_Vector(alternate, consequent'LENGTH);
    END IF;
  END fif;
  FUNCTION fif(test : std_logic; consequent : std_logic_vector; alternate : integer)
    RETURN std_logic_vector IS
  BEGIN
    IF '1' = test THEN
      RETURN consequent;
    ELSE
      RETURN Conv_Std_Logic_Vector(alternate, consequent'LENGTH);
    END IF;
  END fif;

END logical;
