;;;; "models.scm" utility models for synchronous simulation.	-*-scheme-*-
;;; Copyright (C) 1997, 1998 Aubrey Jaffer

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

(require 'array-for-each)
(require 'logical)

(define (integer->prototype obj)
  (cond ((not obj) '#(0))
	((array? obj) obj)
	((char? obj) (string obj))
	((and (integer? obj) (exact? obj) (not (negative? obj)))
	 (let ((wid (integer-length obj)))
	   (cond ((< 64 wid) (vector obj))
		 ((< 32 wid) (A:fixN64b obj))
		 ((< 16 wid) (A:fixN32b obj))
		 ((< 8 wid) (A:fixN16b obj))
		 (else (A:fixN8b obj)))))
	(else (slib:error 'integer->prototype obj))))

(define (make-ram-array prototype . bounds)
  (let ((ara (apply make-array prototype bounds)))
    (array-fill! ara (integer->prototype prototype))
    ara))

(define (make-ram length . prototype)
  (set! prototype (if (null? prototype) #f (car prototype)))
  (make-array (integer->prototype prototype) length))

(define make-asyn-ram make-ram)
(define make-asyn-ram-array make-ram-array)
(define make-sync-ram make-ram)
(define make-sync-ram-array make-ram-array)

(define (make-fifo length . prototype)
  (set! prototype (if (null? prototype) #f (car prototype)))
  (make-array (integer->prototype prototype) (list -2 (+ -1 length))))
(define (fifo:clear fifo)
  (array-fill! fifo 0)
  fifo)
(define (fifo:first fifo)
  (logaref fifo (modulo (logaref fifo -2)
			  (+ 1 (cadar (array-shape fifo))))))
(define (fifo:remove-first fifo)
  (if (fifo:empty? fifo) (synch:error 'fifo-underflow fifo))
  (array-set fifo
	     (modulo (+ 1 (logaref fifo -2))
		     (* 2 (+ 1 (cadar (array-shape fifo)))))
	     -2))
(define (fifo:insert-last fifo datum)
  (if (fifo:full? fifo) (synch:error 'fifo-overflow fifo))
  (array-set (array-set fifo datum (modulo (logaref fifo -1)
					   (+ 1 (cadar (array-shape fifo)))))
	     (modulo (+ 1 (logaref fifo -1))
		     (* 2 (+ 1 (cadar (array-shape fifo)))))
	     -1))
(define (fifo:empty? fifo)
  (= (logaref fifo -1) (logaref fifo -2)))
(define (fifo:full? fifo)
  (define len (+ 1 (cadar (array-shape fifo))))
  (= len (modulo (- (logaref fifo -1) (logaref fifo -2))
		 (* 2 len))))
(define (fifo:fullness fifo)
  (modulo (- (logaref fifo -1) (logaref fifo -2))
	  (* 2 (+ 1 (cadar (array-shape fifo))))))

(define (bounds->prototype bounds)
  (if bounds
      (+ -1 (ash 1 (abs (- (cadr bounds) (caddr bounds)))))
      #f))
(define (must-be-id name)
  (if (not (and (symbol? name)
		(ptag-prefix name)))
      (slib:warn 'synch:fifo name 'not 'id)))

(defmacro synch:fifo (name flen data-source write-strobe read-strobe
			   data-first fullness empty? full?)
  (define mlen (ash 1 (+ -1 (integer-length (+ 1 flen)))))
  (define clen (ash 1 (integer-length flen)))
  ;;(print 'flen '= flen 'mlen '= mlen 'clen '= clen)
  (must-be-id name)
  (and empty? (must-be-id empty?))
  (and full? (must-be-id full?))
  (let* ((bounds (vector-bounds name))
	 (name-no-pre (ptag-stripped name))
	 (pre (symbol-append (ptag-prefix name) ":"))
	 (nam (ptag-stripped (if bounds (car bounds) name)))
	 (idxbits (symbol-append "[" (+ -2 (integer-length clen)) ":0]"))
	 (outidx (symbol-append nam '-outidx idxbits))
	 (inidx (symbol-append nam '-inidx idxbits)))
    (if (not fullness)
	(set! fullness (symbol-append (if bounds (car bounds) name)
				      '-fullness idxbits)))
    (must-be-id fullness)
    ;;(if (vector-bounds first) (slib:warn 'synch:fifo first 'should-not-have-bounds))
    `(begin
       (comment "RAM implementing the FIFO.")
       (synch:define ,pre ,name-no-pre
		     (make-ram ,mlen ,(bounds->prototype bounds))
		     #f #f
		     (if ,write-strobe
			 (array-set ,name-no-pre
				    ,data-source
				    (modulo ,inidx ,mlen))
			 ,name-no-pre))
       (comment "(circular) pointer to word to write.")
       (synch:define ,pre ,outidx 0 #f #f
		     (if ,read-strobe
			 (modulo (+ 1 ,outidx) ,clen)
			 ,outidx))
       (comment "(circular) pointer to word to read.")
       (synch:define ,pre ,inidx 0 #f #f
		     (if ,write-strobe
			 (modulo (+ 1 ,inidx) ,clen)
			 ,inidx))
       (comment "Number of words in use.")
       (synch:defmacro ,fullness (modulo (- ,inidx ,outidx) ,clen))
       (synch:insert-check
	'(cond ((and ,read-strobe (zero? ,(ptag-stripped fullness)))
		(synch:error ',name-no-pre 'reading-while-empty))
	       ((and ,write-strobe (<= ,flen ,(ptag-stripped fullness)))
		(synch:error ',name-no-pre 'writing-while-full))))
       ,(cond (full?
	       (comment "FIFO is full.")
	       `(synch:defmacro ,full? (= ,flen ,(ptag-stripped fullness))))
	      (else #f))
       ,(cond (empty?
	       (comment "FIFO is empty.")
	       `(synch:defmacro ,empty? (zero? ,(ptag-stripped fullness))))
	      (else #f))
       (comment "First-out word.")
       (synch:defmacro ,data-first
		       (logaref ,name-no-pre (modulo ,outidx ,mlen))))))
