;;;; scm2vhdl.scm: Program for translating SIMSYNCH code to VHDL -*-scheme-*-
;;; Copyright (C) 1991, 1992, 1993, 1995, 1996, 1997, 1998, 2000, 2001 Aubrey Jaffer.
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

;;(require 'common-list-functions)
(require 'chapter-order)
(require 'string-search)
(require 'fluid-let)
(require 'printf)
;;(require 'debug) (set! *qp-width* 100) (define qreport qpn)

;;; REPORT an error or warning
(define report
  (lambda args
    (display *vhdl-output-name*)
    (cond (*current-signal-name*
	   (display ": In function `")
	   (display *current-signal-name*)
	   (display "': "))
	  (else (display ": ")))
    (display *vhdl-output-line-number*)
    (display ": warning: ")
    (apply qreport args)))

(define qreport
  (lambda args
    (for-each (lambda (x) (write x) (display #\space)) args)
    (newline)))

;;; This allows us to test without generating files
(define *vhdl-output* (current-output-port))
(define *vhdl-output-name* "?")

;;; *index* is the bit of an integer -- not a subscript.
(define *index* #f)			;dynamic variable for looping vects.
(define *range* #f)			;dynamic variable for numeric constants.
(define *prefix-negation* #f)		;n_ vs. _n
(define *current-signal-name* #f)
(define *vhdl-output-line-number* 0)
(define CONTLINE -80)
(define QEVENT (string->symbol "'EVENT"))

(define (indent->string indent)
  (string-append (make-string (quotient indent 8) #\tab)
		 (make-string (modulo indent 8) #\space)))

;;; OUT indents and displays the arguments
(define (out indent . args)
  (cond ((not (number? indent))
	 (set! args (cons indent args))
	 (set! indent 0)))
  (cond ((>= indent 0)
	 (newline *vhdl-output*)
	 (set! *vhdl-output-line-number* (+ 1 *vhdl-output-line-number*))
	 (display (indent->string indent) *vhdl-output*)))
  (cond ((null? args))
	((or (not (string? (car args)))
	     (not (string-index (car args) #\%)))
	 (for-each (lambda (a)
		     (cond ((eq? '? a) (display "?" *vhdl-output*))
			   ((symbol? a)
			    (translate-name a *vhdl-output*))
			   (else (display a *vhdl-output*))))
		   args))
	(else (apply fprintf *vhdl-output* args))))

(define (translate-name prname port)
  (define visible? #f)
  (define bounds-spec? #f)
  (define pin-rec ((pin-basename-table 'row:retrieve*)
		   (lambda (ptag)
		     (case ptag
		       ((int) #f)
		       (else #t)))
		   (cond ((vector-bounds prname) => car)
			 (else prname))))
  (define nstr (symbol->string prname))
  (case (length pin-rec)
    ((0) #f)
    ((1)
     (let ((ptag (caar pin-rec))
	   (pins (caddar pin-rec)))
       (set! pin-rec (pin-table:row-retrieve ptag (car pins))))
     ;;(print 'pin-rec pin-rec)
     (set! nstr
	   (case (car (cddddr pin-rec))
	     ((#f #t) nstr)		;(string-upcase )
	     (else
	      (set! pin-rec ((signal-table 'get* 'defining-ptag) #f (cadr pin-rec)))
	      (case (length pin-rec)
		((0) #f)
		((1)
		 (display (car pin-rec) port)
		 (display #\_ port))
		(else (report "Multiple defining-ptags for name" prname pin-rec)))
	      nstr)
	     )))
    ;;(else (report "Multiple pins for name" pin-rec) #f)
    )
  (case (and *prefix-negation* (string-ref nstr (+ -1 (string-length nstr))))
    ((#\_ #\-)
     (display "n_" port)
     (set! nstr (substring nstr 0 (+ -1 (string-length nstr))))))
  (for-each
   (lambda (c)
     (let ((tc (cond ((char-alphabetic? c) c)
		     ((char-numeric? c) c)
		     ((char=? c #\.)
		      (cond (bounds-spec? (set! bounds-spec? #f) #\:)
			    (else #f)))
		     ((char=? c #\[) (set! bounds-spec? #t) #\()
		     ((char=? c #\]) (set! bounds-spec? #f) #\))
		     ((char=? c #\') #\')
		     ((char=? c #\-) #\_)
		     ((char=? c #\_) #\_)
		     ((char=? c #\?) "_P")
		     ((char=? c #\%) "Percent")
		     ((char=? c #\:) (if bounds-spec? " DOWNTO " #\_))
		     (else #f))))
       (cond (tc (set! visible? #t)
		 (display tc port)))))
   (string->list nstr))
  (case (and (not *prefix-negation*)
	     (string-ref nstr (+ -1 (string-length nstr))))
    ((#\_ #\-)
     (display #\n port)))
  (if (not visible?) (report "VHDL-invisible symbol?" prname)))

(define (out-comment indent str)
  (out indent "--"
       (string-subst str "\n" (string-append "\n" (indent->string indent)
					     "--"))))

(define (vhdl:revectorize sym ind1 ind2)
  (revectorize ":" sym ind1 ind2))

(define (descmfilify file)
  (let ((sl (string-length file)))
    (cond ((< sl 4) file)
	  ((string-ci=? (substring file (- sl 4) sl) ".scm")
	   (substring file 0 (- sl 4)))
	  (else file))))

(define (numbers-in-expression? tree)
  (let walk ((tree tree))
    (cond ((number? tree))
	  ((not (pair? tree)) #f)
	  ((not (list? tree)) #f)
	  ((memq (car tree) '(logbit? =)) #f) ; ash bit-extract
	  ((some number? (cdr tree)))
	  (else (some walk (cdr tree))))))

(define (registered-input? name)
  (and (symbol? name)
       (let ((name (symbol->string name)))
	 (and (> (string-length name) 4)
	      (string=? "pre:" (substring name 0 4))
	      (string->symbol (substring name 4 (string-length name)))))))

(define (generate-file-header design tag)
  (require 'posix-time)
  (if (file-exists? "Copyrite.vhd")
      (call-with-input-file "Copyrite.vhd"
	(lambda (inport)
	  (do ((line (read-line inport) (read-line inport))
	       (indent CONTLINE 0))
	      ((eof-object? line))
	    (out indent line)))))
  (out 0)
  (let* ((ptag (or (ptag->pld tag) "")) ;"PALCE630"
	 (signature (ptag->sig tag)))
    ;;(out "--TITLE     %s" (symbol-append design '- tag))
    (out "--FILE      %s" *vhdl-output-name*)
    (out "--ENTITY    %s" tag)
    (out "--PROJECT   %s" (or time-stamp:configuration ""))
    (out "--REVISION  %a" time-stamp:revision)
    (and time-stamp:company
	 (out "--COMPANY   %s" time-stamp:company))
    (and time-stamp:author
	 (out "--AUTHOR    %s" time-stamp:author))
    (out "--DATE      %s" (ctime (or time-stamp:time (current-time))))
    (cond ((and signature
		(> (string-length ptag) 4)
		(string-ci=? (substring ptag 0 4) "mach"))
	   (if (not (eqv? signature (logand #xffffffff signature)))
	       (report "signature value out of range"
		       (number->string signature 16))
	       (out "--SIGNATURE #h%x" signature))))
    (if (not (equal? "" ptag)) (out "--PROCESS   %s" ptag))
    (out "")))

(define (out-ports indent type args coms)
  (define (out-bounds fbnds)
    (define bnds fbnds)
    (out CONTLINE ": " type (if (pair? bnds) " std_logic_vector"  " std_logic"))
    (if (and (pair? bnds) (car bnds))
	(out CONTLINE "(" (car bnds) " DOWNTO " (cadr bnds) ")"))
    (cond ((and (pair? bnds) (pair? (cddr bnds)))
	   (out CONTLINE " (")
	   (translate-exp #f 5 (cadddr bnds))
	   (out CONTLINE " DOWNTO ")
	   (translate-exp #f 5 (car (last-pair fbnds)))
	   (out CONTLINE #\)))))
  (if (pair? args)
      (let loop ((args args)
		 (iarg 1)
		 (last-bnds #f)
		 (coms (or coms (map (lambda (arg) #f) args))))
	(if (null? args)
	    (out-bounds last-bnds)
	    (apply
	     (lambda (nm . fbnds)
	       (define bnds (butlast fbnds 1))
	       (cond ((car coms)
		      (if last-bnds (out-bounds last-bnds))
		      (and last-bnds (out CONTLINE ";"))
		      (out-comment indent (car coms))
		      (out indent nm)
		      (loop (cdr args) 1 bnds (cdr coms)))
		     ((equal? bnds last-bnds)
		      (out CONTLINE #\,)
		      (cond ((zero? (modulo iarg 6))
			     (out 10 nm))
			    (else (out CONTLINE #\space nm)))
		      (loop (cdr args) (+ 1 iarg) bnds (cdr coms)))
		     (else
		      (if last-bnds (out-bounds last-bnds))
		      (and last-bnds (out CONTLINE ";"))
		      (out indent nm)
		      (loop (cdr args) 1 bnds (cdr coms)))))
	     (car args))))))

(define *genmem* 100)
(define (out-decls type args)
  (define last-bnds #f)
  (define (but3rd lst)
    (append (butnthcdr 2 lst) (nthcdr 3 lst)))
  (define (out-array-types bnds)
    (define bnd3 (but3rd bnds))
    (cond ((and (equal? bnd3 last-bnds)))
	  (else
	   (set! *genmem* (+ 1 *genmem*))
	   (out 7 "SUBTYPE memory_bus_%d IS bit_vector (%d DOWNTO %d);"
		*genmem* (car bnds) (cadr bnds))
	   (out 7 "TYPE memory_type_%d IS ARRAY (" *genmem*)
	   (translate-exp #f CONTLINE (cadddr bnds))
	   (out CONTLINE " TO ")
	   (translate-exp #f CONTLINE (car (cddddr bnds)))
	   (out CONTLINE ") OF memory_bus_%d;" *genmem*)
	   (set! last-bnds bnd3))))
  (define (out-bounds bnds)
    (cond ((and (pair? bnds) (pair? (cddr bnds)))
	   (out CONTLINE ": memory_type_" *genmem* ";")
	   (out 0))
	  ((and (pair? bnds) (car bnds))
	   (out CONTLINE ": " (if (pair? bnds) " std_logic_vector" " std_logic"))
	   (out CONTLINE "(" (car bnds) " DOWNTO " (cadr bnds) ");"))
	  (else
	   (out CONTLINE ": " (if (pair? bnds) " std_logic_vector" " std_logic")
		";"))))
  (if (pair? args)
      (let loop ((args args)
		 (iarg 1)
		 (last-bnds #f)
		 (coms (map signal->comment args)))
	(if (null? args)
	    (out-bounds last-bnds)
	    (apply
	     (lambda (nm . fbnds)
	       (define bnds (butlast fbnds 1))
	       (cond ((car coms)
		      (if last-bnds (out-bounds last-bnds))
		      (if (and (pair? bnds) (pair? (cddr bnds)))
			  (out-array-types fbnds))
		      (out 0)
		      (out-comment 3 (car coms))
		      (out 3 type " " nm)
		      (loop (cdr args) 1 bnds (cdr coms)))
		     ((equal? bnds last-bnds)
		      (out CONTLINE #\,)
		      (cond ((zero? (modulo iarg 6))
			     (out 10 nm))
			    (else (out CONTLINE #\space nm)))
		      (loop (cdr args) (+ 1 iarg) bnds (cdr coms)))
		     (else
		      (if last-bnds (out-bounds last-bnds))
		      (if (and (pair? bnds) (pair? (cddr bnds)))
			  (out-array-types fbnds))
		      (out 3 type " " nm)
		      (loop (cdr args) 1 bnds (cdr coms)))))
	     (car args))))))

(define (translate-constant-definition var def optcom)
  (cond
   ((eq? '*xxxx* var))
   (else
    (set! *current-signal-name* var)
    (cond (optcom (out 0) (out-comment 3 optcom)))
    (out 3 "CONSTANT " var ": ")
    (cond ((boolean? def)
	   (out CONTLINE "BIT := "(if def "'1'" "'0'")))
	  ((and (number? def) (negative? def))
	   (slib:error 'translating '*xxxx*))
	  (else
	   (let ((len (max 2 (integer-length def))))
	     (out CONTLINE "integer" " := ")
	     (translate-exp #f 8 def))))
    (out CONTLINE ";")
    (set! *current-signal-name* #f))))

(define (set-diff l1 l2)
  (if (null? l2)
      l1
      (reverse
       (let build-difference ((l1 l1)
			      (result '()))
	 (cond ((null? l1)
		result)
	       ((member (car l1) l2) (build-difference (cdr l1) result))
	       (else (build-difference (cdr l1) (cons (car l1) result))))))))

(define (generate-entity design tag)
  (call-with-classified-pins
   tag #f
   (lambda (outputs inouts inputs registers)
     (out-entity design tag inputs inouts outputs signal->comment)
     (out-comment 3 "Clock to Output times for registers in this block")
     (out 3 "CONSTANT  tCQ: TIME := %Ks;" 2e-9)
     (cond (tag
	    ((constant-table 'for-each-row-in-order)
	     (lambda (constant-row)
	       (apply
		(lambda (name value optcom)
		  (cond ((and (not (used-in-ptag:get-ptag name tag))
			      (not (boolean? value))))
			((or (boolean? value) (number? value))
			 (translate-constant-definition name value optcom))))
		constant-row)))))
     (out 0)
     (cond (tag
	    (out 0)
	    (out-comment 0 "Registers") (out 0)
	    (let* ((outs (append inouts outputs))
		   (regs (cond ((null? outs) registers)
			       (else (set-diff registers outs)))))
	      (out-decls "SIGNAL" regs))
	    (let ((wire-lst '()))
	      (signal-table:for-each-row
	       (lambda (signal-row)
		 (apply
		  (lambda (block name defining-ptag signal-type
				 initial-state next-function order)
		    (and (or (not defining-ptag) (eq? tag defining-ptag))
			 (case signal-type
			   ((macro)
			    (let ((bnds (get-bounds name)))
			      (cond ((not (assv (if (pair? bnds)
						    (car bnds) name)
						outputs))
				     (set! wire-lst (cons bnds wire-lst)))))))))
		  signal-row)))
	      (out 0)
	      (out-comment 0 "Combinatorials")
	      (out-decls "SIGNAL" wire-lst)
	      )))
     inouts)))

(define (generate-module-footer design tag block)
  (out 0)
  (out 0 "END model; -- " (or tag design block))
  (out 0))

(define (generate-standard-includes)
  (out "")
  (out "LIBRARY ieee;")
  (out "")
  ;;(out "--USE ieee.numeric_std.all;")
  (out "USE ieee.std_logic_1164.all;")
  (out "USE ieee.std_logic_arith.all;")
  (out "USE ieee.std_logic_unsigned.all;")
  (out "")
  (out "USE work.logical.all;")
  (out-comment 0 "
 The (SIMSYNCH) file \"logical.vhd\" provides several utility functions
 which may be used in this file:

 FIF(test, consequent, alternate) is Functional-IF,
	useful in contexts where VHDL disallows IF statements.

	If TEST is True or '1', then return CONSEQUENT;
	otherwise return ALTERNATE.

 LOGBIT_P(index, operand)
	returns True or '1' if OPERAND(INDEX) is '1';
	otherwise returns False or '0'.

 BIT_FIELD(operand, istart, iend)
	returns OPERAND(IEND downto ISTART).

 RSHIFT(operand, icnt)
	shifts (unsigned) OPERAND ICNT bits to the right;
	The low-order bits are truncated.
 LSHIFT(operand, icnt)
	shifts OPERAND ICNT bits to the left;
	The low-order bits are padded with '0'.
")
  ;;(out "USE work.asyn_ram_256x8.all;") (out "")
  )

(define (make-bit-vector bnds chr)
  (make-string (if bnds (+ 1 (abs (apply - (cdr bnds)))) 1) chr))

(define (out-bus paired-input output-enable name)
  (define bnds (vector-bounds paired-input))
  (out-assignment paired-input bnds 3 #f
		  `(if ,output-enable
		       ,name
		       ,(if bnds (make-bit-vector bnds #\Z) #\Z))))

(define (out-busses lift? *tris*)
  (define busses '())
  (for-each (lambda (bus-spec)
	      (if (eqv? lift? (car bus-spec))
		  (set! busses (cons (cdr bus-spec) busses))))
	    *tris*)
  #+f (let ((paireds '()))
	(for-each
	 (lambda (row)
	   (apply (lambda (paired-input output-enable name)
		    (set! paireds (merge-bounds (get-bounds name) paireds)))
		  row))
	 busses)
	(out-decls "SIGNAL" paireds) (out 0))
  (out 0)
  (cond ((not (null? busses))
	 (out 0)
	 (out-comment 3 "Busses")
	 (for-each
	  (lambda (row)
	    (apply out-bus row))
	  busses))))

(define-syntax debind:if
  (syntax-rules ()
    ((_ (?lit ...) ?pat ?obj ?conseq ?alt)
     (letrec-syntax ((debind1
		      (syntax-rules (?lit ...)
			((_ (??first . ??rest) ??obj ??conseq ??alt)
			 (if (pair? ??obj)
			     (let ((carv (car ??obj))
				   (cdrv (cdr ??obj)))
			       (debind1 ??first carv
					(debind1 ??rest cdrv ??conseq ??alt)
					??alt))
			     ??alt))
			((_ ?lit ??obj ??conseq ??alt)
			 (if (eq? '?lit ??obj) ??conseq ??alt))
			...
			((_ () ??obj ??conseq ??alt)
			 (if (null? ??obj) ??conseq ??alt))
			((_ ??name ??obj ??conseq ??alt)
			 (let ((??name ??obj)) ??conseq)))))
       (debind1 ?pat ?obj ?conseq ?alt)))))

(define-syntax destructuring-case
  (syntax-rules ()
    ((_ "RECUR" ?obj ?lits (?pat ?body1 ?body ...))
     (debind:if ?lits ?pat ?obj (begin ?body1 ?body ...) #f))
    ((_ "RECUR" ?obj ?lits (?pat ?body1 ?body ...) ?clause ...)
     (debind:if ?lits ?pat ?obj (begin ?body1 ?body ...)
		(destructuring-case "RECUR" ?obj ?lits ?clause ...)))
    ((_   ?obj (?lit ...)
	  (?pat ?body1 ?body ...) ...)
     (let ((obj ?obj))
       (destructuring-case "RECUR" obj (?lit ...)
			   (?pat ?body1 ?body ...) ...)))))

(define (process-rams)
  (define (make-name base suffix bnds)
    (string->symbol
     (sprintf #f "%s-%s[%d:%d]"
	      (symbol->string base) (symbol->string suffix)
	      (car bnds) (cadr bnds))))
  (define (parse-ram-next ram-name sexp)
    (define (err . objs)
      (apply report "asyn-ram: unexpected form for next-function:"
	     ram-name sexp objs))
    (define (return name1 name2 write-enable data address)
      (if (eq? ram-name name1 name2)
	  (list write-enable address data)
	  (err)))
    ;;(pprint sexp)
    (destructuring-case
     sexp
     (IF ARRAY-SET)
     ((IF test (ARRAY-SET name1 data address) name2)
      (return name1 name2 test data address))
     ((IF test name2 (ARRAY-SET name1 data address))
      (return name1 name2 `(not ,test) data address))
     (fail (err))))
  ((ram-table 'for-each-row-in-order)
   (lambda (ram-row)
     (destructuring-bind
      (block name output write-enable address data)
      (print 'process-ram ram-row)
      (destructuring-bind
       (block name defining-ptag signal-type initial-state next-function order)
       (signal-table:row-retrieve block name)
       (fluid-let ((*block* block))
	 (let ((bnds (vector-bounds name)))
	   (cond ((and (pair? bnds)
		       (equal? '(7 0) (cdr bnds)))
		  (let ((output (make-name (car bnds) 'output (cdr bnds)))
			(parsed (parse-ram-next name next-function)))
		    (synch:define-signal output defining-ptag 'macro #f #f)
		    (ram-table:row-update
		     (append (list block name output) parsed)))
		  (let ((pin-row (pin-table:row-retrieve defining-ptag name)))
		    (set-car! (cddr pin-row) 'library-module) ;pin-type
		    ((pin-table 'row:update) pin-row)))
		 (else
		  (report "unsupported asyn-ram width" block name))))))))))

(define (named-instantiate module name formals bools actuals)
  (out 0)
  (out 3 name ": " module " PORT MAP(")
  (letrec ((outarg (lambda (formal bool? actual)
		     (out 20 formal)
		     (out CONTLINE " => ")
		     (translate-exp bool? CONTLINE actual))))
    (outarg (car formals) (car bools) (car actuals))
    (for-each (lambda (formal bool? actual)
		(out CONTLINE ",")
		(outarg formal bool? actual))
	      (cdr formals)
	      (cdr bools)
	      (cdr actuals)))
  (out 20 ");"))

(define (instantiate-sync-ram block name)
  (destructuring-bind
   (block name . rest)
   (ram-table:row-retrieve block name)
   (let ((bnds (vector-bounds name))
	 (rabnds (array-bounds block name)))
     (named-instantiate
      (sprintf #f "sram%dkx%d" (car rabnds) (integer-length (cadr rabnds)))
      (car bnds)
      '("DO" "WE" "Addr" "DI")
      '(#f  #t #f  #f)
      rest))))

(define (call-with-vhdl-output name proc)
  (define ofile (string-append name ".vhd"))
  (call-with-output-file ofile
    (lambda (port)
      (printf "generating %s\n" ofile)
      (fluid-let ((*vhdl-output* port)
		  (*vhdl-output-name* ofile)
		  (*vhdl-output-line-number* 0))
	(proc)))))

(define (not-ram? initial)
  (and (pair? initial)
       (not (memq (car initial)
		  '(MAKE-RAM MAKE-RAM-ARRAY MAKE-ASYN-RAM MAKE-FIFO)))))

;;; TRANSLATE-TAG - translate top level forms.
(define (translate-tag design tag)
  (define *tris* '())
  (signal-table:for-each-row
   (lambda (signal-row)
     (apply
      (lambda (block name defining-ptag signal-type
		     initial-state next-function order)
	;;(print 'fluid-let '*block* *block* 'block block signal-row)
	(define block-controls ((block-table 'row:retrieve) block))
	(cond
	 (block-controls
	  (apply
	   (lambda (block-name clock reset clock-frequency synchronous-reset? opt-com)
	     (fluid-let ((*block* block))
	       (define optcom (signal->comment name))
	       (define initial (ceval initial-state))
	       (and
		(or (not defining-ptag) (eq? tag defining-ptag))
		(or (pin-table:row-retrieve tag name)
		    (eq? signal-type 'macro))
		(case signal-type
		  ((macro)
		   (cond ((or initial next-function)
			  (cond (optcom (out 0) (out-comment 1 optcom)))
			  ;;(translate-assignment name 1 next-function)
			  (out-assignment name (vector-bounds name) 1 #f next-function)
			  )))
		  ((registered registered-input combinatorial)
		   (let ((bnds (vector-bounds name))
			 (flags (synch:signal-flags block name)))
		     (cond ((string-index
			     (symbol->string (if bnds (car bnds) name)) #\:))
			   ((eqv? *xxxx* initial))
			   ((eqv? '*xxxx* initial)
			    (report name 'initialized initial-state '?))
			   ((and (boolean? initial) bnds)
			    (report
			     name 'vector 'initialized initial-state '?))
			   ((and (number? initial) (not bnds))
			    (report
			     name 'vector? 'initialized initial-state)))
		     ;;(cond ((memq 'NO-RTL-SIMULATION flags) (out 0 "`ifdef SIMULATE") (out 0 "`else")))
		     (apply
		      (lambda (pin-type pin-number output-enable paired-input defining-block)
			(cond ((and (eq? paired-input name)
				    (not (boolean? output-enable)))
			       ;;(set! paired-input (symbol-append 'pin- paired-input))
			       (synch:warn 'gensym? name)))
			(cond
			 ((and paired-input
			       (not (eq? paired-input #t))
			       (not (boolean? output-enable)))
			  ;;(print 'assign paired-input output-enable name)
			  (let ((lift? (or (provided? 'LIFT-BUS)
					   (memq 'LIFT-BUS flags))))
			    (set! *tris* (cons (list (and lift? #t)
						     paired-input
						     output-enable
						     name)
					       *tris*))
			    (cond
			     ((not lift?)
			      (cond (optcom (out 0) (out-comment 3 optcom)))
			      (out-bus paired-input output-enable name))))))
			(cond
			 ((eq? paired-input #t))
			 ((or
			   (and (number? initial) (not (eqv? *xxxx* initial)))
			   (boolean? initial)
			   (not-ram? initial))
			  (cond (optcom (out 0) (out-comment 1 optcom)))
			  (out 1 "PROCESS (" clock)
			  (if (not (or synchronous-reset?
				       (memq 'synchronous-reset flags)))
			      (out CONTLINE ", " reset))
			  (out CONTLINE ")")
			  (out 1 "BEGIN")
			  (translate-top
			   (if (or synchronous-reset?
				   (memq 'synchronous-reset flags))
			       `(define ,name
				  (if (and ,(symbol-append clock QEVENT)
					   ,clock)
				      (if ,(if (negated? reset)
					       `(not ,reset)
					       reset)
					  ,initial-state
					  ,next-function)))
			       `(define ,name
				  (if ,(if (negated? reset)
					   `(not ,reset)
					   reset)
				      ,initial-state
				      (if (and ,(symbol-append clock QEVENT)
					       ,clock)
					  ,next-function)))))
			  (out 1 "END PROCESS;"))
			 ((and (pair? initial)
			       (eq? (car initial) 'MAKE-SYNC-RAM))
			  (cond (optcom (out 0) (out-comment 3 optcom)))
			  (instantiate-sync-ram block name))
			 (else
			  (cond (optcom (out 0) (out-comment 1 optcom)))
			  #+f
			  (if (not (or (eqv? *xxxx* initial)
				       (equal? name initial)
				       (equal? name reset)))
			      (report "initial-value for" name ':
				      initial-state))
			  (out 1 "PROCESS (" clock ")")
			  ;;(qreport 'non-reset 'register: name)
			  (out 1 "BEGIN")
			  (translate-top
			   `(define ,name
			      (if (and ,(symbol-append clock QEVENT)
				       ,clock)
				  ,next-function)))
			  (out 1 "END PROCESS;"))))
		      (cddr (pin-table:row-retrieve tag name)))
		     ;;(cond ((memq 'NO-RTL-SIMULATION flags) (out 0 "`endif --not SIMULATE")))
		     ))))))
	   block-controls))))
      signal-row)))
  *tris*)

;;; TRANSLATE ptags in design.
(define (translate-design design)
  (define *tris* '())
  (set! *prefix-negation* (provided? 'prefix-negation))
  (process-rams)
  (generate-spew-lists)
  (for-each-ptag
   (lambda (tag pld-type signature optcom)
     (printf "translating %s for ptag %s\n" design tag)
     (synch:check-flags '(SYNCHRONOUS-RESET NO-RTL-SIMULATION LIFT-BUS))
     (call-with-vhdl-output
      (string-append (symbol->string design) "_" (symbol->string tag))
      (lambda ()
	(generate-file-header design tag)
	(generate-standard-includes)
        (cond (optcom (out 0) (out-comment 0 optcom)))
	(generate-entity design tag)
	(out 0)
	(out 0 "BEGIN")
	(set! *tris* (append (translate-tag design tag) *tris*))
	(generate-module-footer design tag #f)))))
  (generate-top-file design (remove-duplicates *tris*)))

(define (generate-top-file design *tris*)
  (if (pair? *dump-list*)
      ;; _stim.vhd supplies stimulus to main entity.
      (call-with-vhdl-output
       (string-append (symbol->string design) "_stim")
       (lambda ()
	 (generate-file-header design 'stim)
	 (generate-standard-includes)
	 (out "USE std.textio.all;")
	 (out "USE work.readtext.all;")
	 (out-comment 0 "
\"readtext.vhd\" extends the std.textio function READ to std_ulogic_vector.
")
	 (out "ENTITY tb IS END tb;")
	 (out "")
	 (out "ARCHITECTURE stimulus OF tb IS")
	 (out 0)
	 (generate-stimulus-file design *tris*)
	 (out "END stimulus;")
	 (out 0)))

      ;;design.vhd (main) file ties chips (ptags) together.
      (call-with-vhdl-output
       (symbol->string design)
       (lambda ()
	 (generate-file-header design 'main)
	 (generate-standard-includes)
	 (print '(generate-entity design #f) (generate-entity design #f)) ;returns *inouts*
	 (generate-main-file design *tris*)
	 (out 0) (out 0 "END model;") (out 0)))))

(define (generate-main-file design *tris*)
  (out-components design)
  (out 0 "BEGIN")
  (out-instantiations design)
  (out-busses #t *tris*))

(define (generate-stimulus-file design *tris*)
  (out-components design)
  (call-with-classified-pins
   #f #f
   (lambda (outputs inouts inputs registers)
     (destructuring-bind
      (filename block frequency clock reset . dump-vars)
      *dump-list*
      (let ((flat-outputs
	     (map (lambda (inp)
		    (if (null? (cdr inp))
			(car inp)
			(caar (last-pair inp))))
		  outputs)))
	(out-decls "SIGNAL" (remove-duplicates (append inputs outputs)))
	(out 0)
	(out 3 "CONSTANT tCQ: TIME := %Ks;" 2e-9)
	(out 0)
	(cond (frequency (out 3 "CONSTANT tCK: TIME := %Ks;" (/ frequency))
			 (out 3 "CONSTANT tCH: TIME := 0.47 * tCK;")
			 (out 3 "CONSTANT tCL: TIME := 0.53 * tCK;")))
	(out 0)
	(out 3 "FILE cmdfile: TEXT IS IN \"%s\";" filename)
	(out 0)
	(out 3 "--\"%s\" line number" filename)
	(out 3 "SIGNAL line_number:  std_logic_vector(11 DOWNTO 0);")
	(out 0)
	(out 0 "BEGIN")
	(out-instantiations design)
	(out-busses #t *tris*)
	(out 0)
	(generate-wave.do frequency clock reset dump-vars)
	(out-stimulus-process frequency clock reset dump-vars flat-outputs))))))

(define (generate-wave.do frequency clock reset dump-vars)
  (call-with-output-file "wave.do"
    (lambda (oport)
      (fprintf oport "onerror {resume}\n")
      (fprintf oport "quietly WaveActivateNextPane {} 0\n")
      (fprintf oport "add wave -format Literal -radix decimal /tb/line_number\n")
      (for-each
       (lambda (sig)
	 (define bnds (vector-bounds sig))
	 (cond
	  (bnds
	   (fprintf oport "add wave -format Literal -radix hexadecimal /tb/")
	   (translate-name (car bnds) oport)
	   (newline oport))
	  (else
	   (fprintf oport "add wave -format Logic /tb/" sig)
	   (translate-name sig oport)
	   (newline oport))))
       (list* reset clock dump-vars))
      (fprintf oport "TreeUpdate [SetDefaultTree]      \n")
      (fprintf oport "WaveRestoreCursors {375 ns}      \n")
      (fprintf oport "WaveRestoreZoom {0 ns} {250 ns}  \n")
      (fprintf oport "configure wave -namecolwidth 150 \n")
      (fprintf oport "configure wave -valuecolwidth 100\n")
      (fprintf oport "configure wave -justifyvalue left\n")
      (fprintf oport "configure wave -signalnamewidth 0\n")
      (fprintf oport "configure wave -snapdistance 10  \n")
      (fprintf oport "configure wave -datasetprefix 0  \n"))))

(define (out-stimulus-process frequency clock reset dump-vars flat-outputs)
  (define (translate-clock clock indent val)
    (out indent clock (if val " <= '1';" " <= '0';")))

  (out 3 "PROCESS")
  (out 6 "VARIABLE line_in: LINE;")
  (out 6 "VARIABLE good: BOOLEAN;")
  (for-each (lambda (signal)
	      (define bounds (vector-bounds signal))
	      (cond (bounds
		     (out 6 "VARIABLE tmp_" (car bounds))
		     (out CONTLINE ": std_ulogic_vector(%d DOWNTO %d);"
			  (cadr bounds) (caddr bounds)))
		    (else
		     (out 6 "VARIABLE tmp_" signal ": std_ulogic;"))))
	    dump-vars)
  (out 0)
  (out 3 "BEGIN")
  (out 0)
  (out 6 reset (if (negated? reset) " <= '0';" " <= '1';"))

  (out 6 "LOOP")
  (out 8 "IF EndFile(cmdfile) THEN EXIT; END IF;")
  (out 8 "ReadLine(cmdfile, line_in);")
  (out 8 "NEXT WHEN line_in'LENGTH = 0;")
  (out-comment 8 "Reset while reading lines starting with '.'")
  (out 8 "IF '.' = line_in(1) AND line_in'LENGTH = 1 THEN EXIT; END IF;")
  (translate-clock clock 8 #t)
  (out 8 "WAIT FOR tCH;")
  (translate-clock clock 8 #f)
  (out 8 "WAIT FOR tCL;")
  ;;(out 8 "line_number <= 1 + line_number;")
  (out 6 "END LOOP;")
  (out 0)

  (translate-clock clock 6 #t)
  (translate-assignment reset 6 (negated? reset))
  (out 6 "WAIT FOR tCH;")
  (translate-clock clock 6 #f)
  (out 6 "WAIT FOR tCL;")
  (out 6 "line_number <= \"000000000000\";")
  (out 0)

  (out 6 "LOOP")
  (out 8 "IF EndFile(cmdfile) THEN EXIT; END IF;")
  (out 8 "ReadLine(cmdfile, line_in);")
  (out 8 "NEXT WHEN line_in'LENGTH = 0;")
  (out 0)
  (translate-clock clock 8 #t)
  (for-each
   (lambda (signal)
     (define bounds (vector-bounds signal))
     (define flat-output? (memq signal flat-outputs))
     (out 0)
     (cond
      (bounds
       (out 10 "Read(line_in, tmp_" (car bounds) ", good);")
       (out 10 (sprintf #f "ASSERT good REPORT \"Reading %s\" SEVERITY FAILURE;"
			signal))
       (cond (flat-output?)
	     (else
	      (out 10 "IF (good) THEN")
	      (out 12 signal " <= To_StdLogicVector(" "tmp_" (car bounds) ")"
		   tcq-semi)
	      (out 10 "END IF;"))))
      (else
       (out 10 "Read(line_in, tmp_" signal ", good);")
       (out 10 (sprintf #f "ASSERT good REPORT \"Reading %s\" SEVERITY FAILURE;"
			signal))
       (cond (flat-output?)
	     (else
	      (out 10 signal " <= To_StdLogic(tmp_" signal ")" tcq-semi))))))
   dump-vars)
  (out 0)
  (out 8 "WAIT FOR tCH;")
  (translate-clock clock 10 #f)
  (out 8 "WAIT FOR tCL;")
  (out 0)
  (out-comment 10 "Check outputs")
  (for-each
   (lambda (signal)
     (define bounds (vector-bounds signal))
     (define flat-output? (memq signal flat-outputs))
     (cond
      (bounds
       (cond (flat-output?
	      (out 10 "ASSERT %#A = " (make-string (bounds->width bounds) #\U))
	      (out CONTLINE "tmp_" (car bounds))
	      (out 16 "OR %#A = " (make-string (bounds->width bounds) #\Z))
	      (out CONTLINE "tmp_" (car bounds))
	      (out 16 "OR "
		   signal " = To_StdLogicVector(" "tmp_" (car bounds) ")")
	      (out 12 (sprintf #f "REPORT \"Mismatched %s\" SEVERITY ERROR;"
			       signal))
	      (out 0))))
      (else
       (cond (flat-output?
	      (out 10 "ASSERT 'U' = tmp_" signal)
	      (out 16 " OR 'Z' = tmp_" signal)
	      (out 16 " OR " signal " = To_StdLogic(tmp_" signal ")")
	      (out 12 (sprintf #f "REPORT \"Mismatched %s\" SEVERITY ERROR;"
			       signal))
	      (out 0))))))
   dump-vars)
  (out 8 "line_number <= 1 + line_number;")
  (out 0)
  (out 6 "END LOOP;")
  (out 6 "WAIT;")
  (out 3 "END PROCESS;")
  (out 0))

(define (out-entity design tag inputs inouts outputs signal->comment)
  (if (not signal->comment) (set! signal->comment (lambda (sig) #f)))
  (out 0)
  (out 0 "ENTITY " (or tag design) " IS PORT (")
  (out-ports 3 "IN" inputs (map signal->comment inputs))
  (if (and (not (null? inputs)) (not (null? outputs))) (out CONTLINE #\;))
  (out-ports 3 "BUFFER" outputs (map signal->comment outputs))
  (cond ((null? inouts))
	(else
	 (out CONTLINE #\;)
	 (out 0)
	 (out-comment 3 "Bidirectional signals")
	 (out-ports 3 "INOUT" inouts (map signal->comment inouts))
	 ))
  (out CONTLINE ");")
  (out 0 "END " (or tag design) ";")
  (out 0)
  (out 0 "ARCHITECTURE model OF " (or tag design) " IS"))

(define (out-component design tag inputs inouts outputs signal->comment)
  (out 3 "COMPONENT " (or tag design) " PORT(")
  (out-ports 6 "IN" inputs (map signal->comment inputs))
  (if (and (not (null? inputs)) (not (null? outputs))) (out CONTLINE #\;))
  (out-ports 6 "BUFFER" outputs (map signal->comment outputs))
  (cond ((null? inouts))
	(else
	 (out CONTLINE #\;)
	 (out 0)
	 (out-comment 6 "Bidirectional signals")
	 (out-ports 6 "INOUT" inouts (map signal->comment inouts))
	 ))
  (out CONTLINE ");")
  (out 3 "END COMPONENT;")
  (out 0))

(define (out-components design)
  (for-each-ptag
   (lambda (tag pld-type signature optcom)
     (call-with-classified-pins
      tag #f
      (lambda (outputs inouts inputs registers)
	(cond (optcom (out-comment 3 optcom)))
	(out-component design tag inputs inouts outputs signal->comment))))))

(define *instance-count* 0)
(define (instantiate design tag inputs inouts outputs)
  (define loop (lambda (i args)
		 (cond ((pair? args)
			(let ((a (car args)))
			  (out (if (zero? (modulo i 3)) 6 CONTLINE)
			       (car a) "=>" (car a))
			  (if (pair? (cdr args))
			      (out CONTLINE ", "))
			  (loop (+ i 1) (cdr args)))))))
  (define outsigs (lambda (name sigs)
		    (cond ((not (null? sigs))
			   (out 12 "--module " name)
			   (loop 0 sigs)))))
  (print 'instantiate design tag)
  (set! *instance-count* (+ 1 *instance-count*))
  (out 3 (or tag design) "_" *instance-count* ": " (or tag design) " PORT MAP(")
  (outsigs "inputs" inputs)
  (if (pair? inouts) (out CONTLINE ","))
  (outsigs "inouts" inouts)
  (if (pair? outputs) (out CONTLINE ","))
  (outsigs "outputs" outputs)
  (out CONTLINE ");"))

(define (out-instantiations design)
  (for-each-ptag
   (lambda (tag pld-type signature optcom)
     (call-with-classified-pins
      tag #f
      (lambda (outputs inouts inputs registers)
	(cond (optcom (out 0) (out-comment 3 optcom)))
	(instantiate design tag inputs inouts outputs))))))

(define (for-each-index proc bounds)
  (let* ((first (cadr bounds))
	 (last (caddr bounds))
	 (range (- last first)))
    (if (positive? range)
	(do ((i 0 (+ 1 i))
	     (j first (+ 1 j)))
	    ((> i range))
	  (fluid-let ((*index* i))
	    ;;(print '+range i j)
	    (proc j)))
	(do ((i (- range) (+ -1 i))
	     (j first (+ -1 j)))
	    ((negative? i))
	  (fluid-let ((*index* i))
	    ;;(print '-range i j)
	    (proc j))))))

(define (for-each-value proc bounds)
  (let* ((first (cadr bounds))
	 (last (caddr bounds))
	 (range (- last first)))
    (if (not (zero? (min first last)))
	(report "mapping over non-consecutive values" bounds))
    (set! range (ash 2 (abs range)))
    (do ((i 0 (+ 1 i)))
	((>= i range))
      (fluid-let ((*index* i))
	(proc i)))))

;;; TRANSLATE-TOP - translate top level form sexp.
(define (translate-top sexp)
  (cond ((symbol? sexp) (set! *current-signal-name* sexp))
	((and (pair? sexp) (eq? (car sexp) 'QUOTE))
	 (set! *current-signal-name* (cadr sexp)))
	((string? sexp) (out-comment 0 sexp) (out 0))
	((not (pair? sexp))
	 (report "top level atom?" sexp))
	(else (case (car sexp)
		((define) (translate-assignment (cadr sexp) 3 (caddr sexp)))
		(else (report "statement not in procedure" sexp)))))
  ;;(out 0)
  (set! *current-signal-name* #f))

(define (case-able? rhs)
  (and (pair? rhs)
       (case (car rhs)
	 (( CASE QASE CASEV IF AND NOT ARRAY-SET LOGASET ) #t) ; NUMBER-OR OR
	 ((SYNCH:PRE) (case-able? (signal-table:get-next-fun *block* (cadr rhs))))
	 (else #f))))

(define (when-able? rhs)
  (and (pair? rhs)
       (case (car rhs)
	 (( IF AND NOT ARRAY-SET LOGASET ) #t)
	 ((SYNCH:PRE) (when-able? (signal-table:get-next-fun *block* (cadr rhs))))
	 (else #f))))

(define (translate-assignment lhs indent rhs)
  (set! *current-signal-name* lhs)
  (translate-ca-assign lhs indent #f rhs))

(define (translate-ca-assign lhs indent pol rhs)
  (cond ((and (equal? lhs rhs)
	      (not pol)
	      (not (null? (array-bounds *block* lhs))))
	 (out CONTLINE " ;"))
	((case-able? rhs)
	 (case (car rhs)
	   ((NOT)
	    (translate-ca-assign lhs indent (not pol) (cadr rhs)))
	   ((AND) (translate-ca-assign
		   lhs
		   indent
		   pol
		   (case (length rhs)
		     ((1) #t)
		     ((2) (cadr rhs))
		     (else `(if ,(butlast rhs 1) ,(car (last rhs 1)) #f)))))
	   ((IF)
	    (translate-assign-if lhs indent pol rhs #t #f))
	   ((CASE QASE CASEV)
	    (let ((elsed? #f))
	      (out indent "CASE ")
	      (integerize (+ 6 indent) (cadr rhs))
	      (out CONTLINE " IS")
	      (for-each (lambda (clause)
			  (if (translate-case-clause lhs (+ 2 indent) pol clause)
			      (set! elsed? #t)))
			(cddr rhs))
	      (cond ((not elsed?)
		     (out (+ 2 indent) "WHEN OTHERS =>")
		     (out (+ 4 indent)
			  "ASSERT false REPORT \"Unexpected CASE\" SEVERITY FAILURE;")))
	      (out indent "END CASE;")))
	   ((SYNCH:PRE)
	    (let ((expn (signal-table:get-next-fun *block* (cadr rhs))))
	      (if (not expn) (report "returned #f" rhs))
	      (translate-ca-assign lhs indent pol expn)))
	   ((ARRAY-SET LOGASET)
	    (if (not (equal? lhs (cadr rhs)))
		(report 'array-set lhs rhs))
	    (let ((bounds (vector-bounds lhs)))
	      (out-assignment (list (if bounds (car bounds) lhs) (cadddr rhs))
			      bounds indent pol (caddr rhs))))
	   (else (report "case-able expression not translated" rhs))))
	;;((when-able? rhs) (translate-assign-when lhs indent pol rhs #t))
	(else (out-assignment lhs (vector-bounds lhs) indent pol rhs))))

(define (integerize indent sexp)
  (out CONTLINE "Conv_Integer(")
  (translate-exp #f (+ 6 indent) (remove-gratuitous-masking sexp #f))
  (out CONTLINE ")"))

(define (invertify sexp)
  (cond ((and (pair? sexp) (eq? 'NOT (car sexp)))
	 (cadr sexp))
	((boolean? sexp) (not sexp))
	(else (list 'NOT sexp))))
(define (spin pol sexp)
  (if pol (invertify sexp) sexp))

(define (remove-gratuitous-masking rhs width)
  (cond ((not (pair? rhs)) rhs)
	((and (eqv? 'MODULO (car rhs))
	      (eqv? (caddr rhs)
		    (ash 1 (or width (vector-width (cadr rhs)) -1))))
	 (cadr rhs))
	((and (eqv? 'LOGAND (car rhs))
	      (eqv? (cadr rhs)
		    (+ -1 (ash 1 (or width (vector-width (caddr rhs)) -1)))))
	 (caddr rhs))
	(else rhs)))
;;(trace remove-gratuitous-masking vector-width integerize)(set! *qp-width* 333)

(define tcq " AFTER tCQ")
(define tcq-semi " AFTER tCQ;")
(define (out-assignment lhs bounds indent pol rhs)
  (define casing? (and (pair? rhs) (memq (car rhs) '(CASE QASE CASEV))))
  (define elsed? #f)
  (define frst? #t)
  (cond (casing?
	 (out indent "WITH ")
	 (integerize (+ 6 indent) (cadr rhs))
	 (out CONTLINE " SELECT")
	 (set! indent (+ 2 indent))))
  (cond ((pair? lhs)
	 (out indent (car lhs) "(")
	 (integerize (+ 10 indent) (cadr lhs))
	 (out CONTLINE ")"))
	(else (out indent lhs)))
  (out CONTLINE " <= ")
  (if (and (pair? rhs) (pair? lhs)) (out (+ 1 indent)))
  (set! indent (+ 2 indent))
  (cond (casing?
	 (for-each
	  (lambda (clause)
	    (cond ((and (pair? clause)
			(eq? 'ELSE (car clause))
			(not (null? (cdr clause))))
		   (if frst? (set! frst? #f) (out CONTLINE ","))
		   (if (translate-select-clause
			(pair? lhs) bounds indent pol clause)
		       (set! elsed? #t)))
		  ((and (pair? clause) (not (null? (cdr clause))))
		   (for-each (lambda (match)
			       (if frst? (set! frst? #f) (out CONTLINE ","))
			       (if (translate-select-clause
				    (pair? lhs) bounds indent pol
				    (cons (list match) (cdr clause)))
				   (set! elsed? #t)))
			     (car clause)))
		  (else
		   (if frst? (set! frst? #f) (out CONTLINE ","))
		   (if (translate-select-clause
			(pair? rhs) bounds (+ 5 indent) pol clause)
		       (set! elsed? #t)))))
	  (cddr rhs))
	 (cond ((not elsed?)
		(out CONTLINE ",")
		(out indent lhs " AFTER tCQ WHEN OTHERS"))))
	(else (out-rhs (pair? lhs) bounds indent pol rhs)))
  (out CONTLINE ";"))

(define (out-rhs lhp? bounds indent pol rhs)
  (cond
   ((and (pair? rhs) (eq? 'NOT (car rhs)))
    (out-rhs lhp? bounds indent (not pol) (cadr rhs)))
   ((and (pair? rhs) (eq? 'IF (car rhs)))
    (out-convert lhp? bounds (+ 2 indent) pol (caddr rhs))
    (out (+ 2 indent) "WHEN ")
    (translate-exp #t CONTLINE (cadr rhs))
    (out (+ 2 indent) "ELSE ")
    (out-rhs lhp? bounds (+ 2 indent) pol (cadddr rhs)))
   (else (out-convert lhp? bounds indent pol rhs))))

(define (out-convert lhp? bounds indent pol rhs)
  (cond (bounds
	 (let ((wid (bounds->width bounds))
	       (rhsym (vector-bounds rhs)))
	   ;;(out 0 "--bounds=" bounds " wid=" wid " rhsym=" rhsym " rhs=" rhs)
	   (cond ((and rhsym (not lhp?) (eqv? (bounds->width rhsym) wid))
		  (translate-exp #f indent
				 (spin pol (remove-gratuitous-masking rhs wid)))
		  (out CONTLINE "" tcq))
		 ((eqv? '*xxxx* rhs)
		  (out CONTLINE "\"" (make-string wid #\X) "\"" tcq))
		 ((number? rhs)
		  (out CONTLINE "\"%0.*b\"" wid rhs) (out CONTLINE tcq))
		 ((and (pair? rhs)
		       (memq (car rhs) '(ARRAY-REF LOGAREF BYTE-REF)))
		  (translate-exp #f indent
				 (spin pol (remove-gratuitous-masking rhs wid)))
		  (out CONTLINE tcq))
		 (else
		  (out CONTLINE
		       (cond (lhp? "To_BitVector")
			     ((and (pair? rhs)
				   (eq? 'BOOLEANS-TO-NUMBER (car rhs)))
			      "Ext")
			     (else "Conv_Std_Logic_Vector"))
		       "(")
		  (translate-exp #f (+ 5 indent)
				 (spin pol (remove-gratuitous-masking rhs wid)))
		  (if (not lhp?) (out CONTLINE ", " wid))
		  (out CONTLINE ")" tcq)))))
	((eqv? '*xxxx* rhs) (out CONTLINE "'X'" tcq))
	((symbol? rhs)
	 (let ((sym (spin pol rhs)))
	   (cond ((pair? sym) (out CONTLINE "NOT " (cadr sym) tcq))
		 (else (out CONTLINE rhs tcq)))))
	((not (boolean? rhs))
	 (out CONTLINE "To_StdLogic(")
	 (translate-exp #t (+ 5 indent) (spin pol rhs))
	 (out CONTLINE ")" tcq))
	((spin pol rhs) (out CONTLINE "'1'" tcq))
	(else (out CONTLINE "'0'" tcq))))

(define (case-neval dat clause)
  (cond ((not (and (pair? dat) (memq (car dat) '(UNQUOTE-SPLICING UNQUOTE))))
	 dat)
	((symbol? (cadr dat)) (cadr dat))
	(else (cadr dat))))

(define (case-eval dat clause)
  (cond ((not (and (pair? dat) (memq (car dat) '(UNQUOTE-SPLICING UNQUOTE))))
	 dat)
	((symbol? (cadr dat)) (cadr dat))
	(else (slib:eval (cadr dat)))))

(define (translate-case-clause lhs indent pol clause)
  (cond ((eq? 'ELSE (car clause))
	 (let ((lc (car (last-pair clause))))
	   (cond ((and (pair? lc)
		       (memq (car lc) '(synch:chat synch:warn synch:error))))
		 ;;((equal? lhs lc))
		 (else
		  ;;(if (not (member lc '(0 #f))) (report "case else not (0) #f?" lc))
		  (out indent "WHEN OTHERS =>")
		  (translate-ca-assign lhs (+ 2 indent) pol lc))))
	 #t)
	((= 1 (length (car clause)))
	 (let ((dat (caar clause)))
	   (out indent "WHEN " (case-eval dat clause) " => "))
	 (translate-ca-assign lhs (+ 2 indent)
			       pol (car (last-pair clause)))
	 #f)
	(else
	 (for-each (lambda (match)
		     (translate-case-clause
		      lhs indent pol (cons (list match) (cdr clause))))
		   (car clause))
	 #f)))

(define (translate-select-clause lhp? bounds indent pol clause)
  (out indent)
  (cond ((eq? 'ELSE (car clause))
	 (let ((lc (car (last-pair clause))))
	   (cond ((and (pair? lc)
		       (memq (car lc) '(synch:chat synch:warn synch:error))))
		 (else
		  ;;(if (not (member lc '(0 #f))) (report "select else not (0) #f?" lc))
		  (out-convert lhp? bounds indent pol lc)
		  (out indent "WHEN OTHERS"))))
	 #t)
	((= 1 (length (car clause)))
	 (out-convert lhp? bounds indent pol (car (last-pair clause)))
	 (let ((dat (caar clause)))
	   (out indent "WHEN ")
	   (translate-exp #f CONTLINE (case-neval dat clause)))
	 #f)
	(else (slib:error 'translate-select-clause (car clause)) #f)))

(define (translate-assign-if lhs indent pol rhs sense tailing?)
  (define test (cadr rhs))
  (define consequent (caddr rhs))
  (define alternate (if (null? (cdddr rhs)) '*no-else* (cadddr rhs)))
  (case (and (pair? test) (car test))
    ((NOT) (translate-assign-if lhs indent pol
				`(if ,(cadr test) ,consequent ,alternate)
				(not sense)
				tailing?))
    ((DEFINED? PROVIDED?)
     (translate-ifdef test consequent alternate
		      (lambda (sexp)
			(if (negative? indent) (set! indent 16))
			(out indent)
			(translate-ca-assign lhs (+ 2 indent) pol sexp))
		      sense))
    (else
     (out (if tailing? CONTLINE indent) "IF ")
     (translate-exp #t CONTLINE (if sense test `(not ,test))) ;(+ 2 indent)
     (out CONTLINE " THEN")
     (translate-ca-assign lhs (+ 2 indent) pol consequent)
     (cond ((null? (cddr rhs)) (report 'dangling-if? lhs rhs)
	    (out indent "ELSE;")
	    (out indent "END IF;"))
	   ((and (pair? alternate) (eqv? 'IF (car alternate)))
	    (out indent "ELS")
	    (translate-assign-if lhs indent pol alternate #t #t))
	   ((and (eqv? '*no-else* alternate)
		 ;;(not (null? (array-bounds *block* lhs)))
		 )
	    (out indent "END IF;"))
	   ((equal? lhs alternate)
	    (out indent "END IF;"))
	   (else
	    (out indent "ELSE")
	    (translate-ca-assign lhs (+ 2 indent) pol alternate)
	    (out indent "END IF;"))))))

;;; LOOKUP - translate from table or return arg as string
(define (lookup arg tab)
  (let* ((p (assq arg tab))
	 (l (if p (cdr p) arg)))
    (if (symbol? l) (symbol->string l) l)))

(define (out-number sexp)
  (if *range*
      (let* ((num (if (positive? *range*)
		      (reverse-bit-field sexp 0 *range*)
		      sexp))
	     (str (number->string num 2)))
	(print ">>>> Time to debug numbers within *range*.")
	(out CONTLINE
	     (number->string (abs *range*))
	     "\""
	     (make-string (- (abs *range*) (string-length str)) #\0)
	     str
	     "\""))
      (out CONTLINE (if (> sexp #x7fffffff) "\"%b\"" "%02d") sexp)))

(define (attribute? sym)
  (substring? "'" (symbol->string sym)))

;;; TRANSLATE-EXP - translate expression
(define (translate-exp bool? indent sexp)
  (cond ((not (pair? sexp))		;atoms
	 (cond ((string? sexp)
		(out CONTLINE #\" sexp #\"))
	       ((char? sexp)
		(out CONTLINE (string #\' sexp #\')))
	       ((boolean? sexp)
		(out CONTLINE
		     (if bool?
			 (case sexp ((#f) "FALSE") ((#t) "TRUE"))
			 (case sexp ((#f) "'0'") ((#t) "'1'")))))
	       ((and *index* (symbol? sexp) (vector-bounds sexp))
		(translate-exp bool? indent (if (negative? *index*)
						#f
						`(logbit? ,*index* ,sexp))))
	       ((and *index* (integer? sexp))
		(translate-exp bool? indent
			       (and (not (negative? *index*))
				    (logbit? *index* sexp))))
	       ((symbol? sexp) (if (and bool? (not (attribute? sexp)))
				   (out CONTLINE "('1' = " sexp ")")
				   (out CONTLINE sexp)))
	       ((number? sexp) (out-number sexp))
	       (else (report "unexpected datatype:" sexp)
		     (out CONTLINE sexp))))
	(else
	 (case (car sexp)
	   ;; These vectorized functions are for multiple bit signals:
	   ((< <= > >=)
	    (infix-translate-pairwise (car sexp) indent (cdr sexp)))
	   ((= EQV?)
	    (infix-translate-pairwise '= indent (cdr sexp)))
	   ((ZERO?)
	    (out CONTLINE "00 = ")
	    (translate-exp #f indent (cadr sexp)))
	   ((POSITIVE?)
	    (out CONTLINE "00 < ")
	    (translate-exp #f indent (cadr sexp)))
	   ((NEGATIVE?)
	    (out CONTLINE "00 > ")
	    (translate-exp #f indent (cadr sexp)))
	   ((LOGTEST)
	    (translate-exp bool? indent `(not (zero? (logand ,@(cdr sexp))))))
	   ((INTEGER->CHAR) (translate-exp #f indent (cadr sexp)))
	   ((LOGBIT?)
	    (cond
	     ((and (number? (cadr sexp)) (negative? (cadr sexp))
		   (translate-exp bool? indent #f)))
	     ((and (number? (cadr sexp)) (symbol? (caddr sexp)))
	      (let* ((bounds (vector-bounds (caddr sexp)))
		     (lowbnd (if bounds (min (cadr bounds) (caddr bounds)) 0))
		     (dvs (devectorize (caddr sexp) (+ lowbnd (cadr sexp)))))
		;;(if (not (zero? lowbnd)) (print 'logbit? sexp bounds lowbnd))
		(cond ((not dvs) (out CONTLINE dvs))
		      (bool? (out CONTLINE "('1' = " dvs
				  "(" (+ lowbnd (cadr sexp)) "))"))
		      (else
		       (out CONTLINE dvs "(" (+ lowbnd (cadr sexp)) ")")))))
	     (else (out CONTLINE 'LOGBIT?)
		   (infix-translate-list #f (+ 2 indent) (cdr sexp)))))
	   ((NOT)
	    (cond ((and bool? (symbol? (cadr sexp)))
		   (out CONTLINE "('0' = " (cadr sexp) ")"))
		  (else
		   (out CONTLINE "NOT ")
		   (optionally-parenthesize bool? (+ 1 indent) (cadr sexp) #f))))
	   ((OR)
	    (cond ((memq #t (cdr sexp)) (translate-exp #t indent #t))
		  (else (infix-translate-exp
			 #t " OR " indent (remove #f (cdr sexp))))))
	   ((XOR)
	    (infix-translate-exp #t " XOR " indent (remove #f (cdr sexp))))
	   ((&&)
	    (infix-translate-exp #t " AND " indent (cdr sexp)))
	   ((AND)
	    (cond ((memq #f (cdr sexp)) (translate-exp bool? indent #f))
		  (else
		   (translate-and bool? indent (remove #t (cdr sexp))))))

	   ((ARRAY-REF LOGAREF BYTE-REF)
	    (translate-logaref indent (cadr sexp) (cddr sexp)))
	   ((NUMBER-OR)
	    (infix-translate-exp #f " OR " indent (cddr sexp)))
	   ((LOGNOT)
	    (cond ((symbol? (cadr sexp))
		   (out CONTLINE "NOT ")
		   (translate-exp bool? (+ 2 indent) (cadr sexp)))
		  (else
		   (out CONTLINE "NOT (")
		   (translate-exp bool? (+ 2 indent) (cadr sexp))
		   (out CONTLINE ")"))))
	   ((LOGIOR LOGAND LOGXOR)
	    (out CONTLINE "(")
	    (infix-translate-arith
	     (lookup (car sexp)
		     '((LOGIOR . " OR ") (LOGAND . " AND ") (LOGXOR . " XOR ")))
	     (+ 5 indent)
	     (cdr sexp))
	    (out CONTLINE ")"))
	   ((- / QUOTIENT + *)		; MODULO
	    (infix-translate-arith
	     (lookup (car sexp)
		     '((- . " - ") (/ . " / ") (+ . " + ") (* . " * ")
		       (MODULO . " MOD ") (QUOTIENT . " / ")))
	     (+ 5 indent)
	     (cdr sexp)))
	   ((MODULO)
	    (translate-exp bool? (+ 2 indent)
			   (if (number? (caddr sexp))
			       `(logand ,(- (caddr sexp) 1)
					,(cadr sexp))
			       `(logand (- ,(caddr sexp) 1)
					,(cadr sexp)))))
	   ((ASH ARITHMETIC-SHIFT)
	    (cond (*index* (fluid-let ((*index* (- *index* (caddr sexp))))
			     (translate-exp bool? indent (cadr sexp))))
		  ((not (number? (caddr sexp)))
		   (slib:warn (car sexp) sexp)
		   (translate-exp #f indent `(SHL ,@(cdr sexp))))
		  (else
		   (translate-exp
		    bool? (+ indent 5)
		    `(,(if (negative? (caddr sexp)) 'rshift 'lshift)
		      ,(cadr sexp) ,(abs (caddr sexp)))))))
	   ((BIT-EXTRACT BIT-FIELD)
	    ;;(print (car sexp) (cdr sexp))
	    (let* ((bounds (vector-bounds (cadr sexp)))
		   (lowbnd (and bounds (min (cadr bounds) (caddr bounds)))))
	      (cond ((not bounds)
		     (report "bit extract of non-symbol?" sexp))
		    ((and (number? (caddr sexp))
			  (number? (cadddr sexp)))
		     (out CONTLINE (vhdl:revectorize
				    (cadr sexp)
				    (+ lowbnd (caddr sexp))
				    (+ lowbnd (cadddr sexp) -1))))
		    (else
		     (out CONTLINE 'bit-field)
		     (infix-translate-list #f (+ 2 indent) (cdr sexp))))))
	   ((TO-STD-LOGIC)
	    (out CONTLINE "To_StdLogic(")
	    (translate-exp bool? (+ 2 indent) (cadr sexp))
	    (out CONTLINE ")"))
	   ((BOOLEANS-TO-NUMBER)
	    (infix-translate-exp #t " & " (+ indent 2)
				 (map (lambda (sxp) `(TO-STD-LOGIC ,sxp))
				      (cdr sexp))))
	   ((SYNCH:PRE)
	    (let ((expn (signal-table:get-next-fun *block* (cadr sexp))))
	      ;;(print 'synch:pre (cadr sexp) '*block* *block* '-> expn)
	      (if (not expn) (report "returned #f" sexp 'in *block*))
	      (translate-exp bool? indent
			     (or expn (symbol-append
				       *block* '- (cadr sexp) '?)))))
	   ((IF)
	    (translate-if bool? indent sexp #t))
	   ((CASE QASE CASEV)
	    (translate-case bool? indent (cadr sexp) (cddr sexp)))
	   ((CHECK=)
	    (translate-exp bool? indent (cadddr sexp)))
	   ((SYNCH:CHAT NUMBER-CHECK)
	    (out CONTLINE "1"))
	   ;;((SYNCH:PRE) (translate-exp bool? indent (macroexpand-1 sexp)))
	   (else (cond ((and (list? sexp) (= 2 (length sexp))
			     (apply signal-table:row-retrieve sexp))
			(translate-exp bool? indent (cadr sexp)))
		       #+f
		       ((pair? (car sexp)) ;computed function
			(out indent "(*(")
			(translate-exp #f (+ 3 indent) (car sexp))
			(out CONTLINE "))")
			(out (+ 2 indent))
			(infix-translate-list
			 #f (+ 2 indent) (cdr sexp)))
		       (else
			(out CONTLINE (car sexp))
			(infix-translate-list
			 bool? (+ 2 indent) (cdr sexp)))))))))

(define (translate-if bool? indent sexp sense)
  (define test (cadr sexp))
  (define consequent (caddr sexp))
  (define alternate (cadddr sexp))
  (case (and (pair? test) (car test))
    ((NOT) (translate-if bool? indent
			 `(if ,(cadr test) ,consequent ,alternate)
			 (not sense)))
    ((DEFINED? PROVIDED?)
     (translate-ifdef test consequent alternate
		      (lambda (sexp)
			(if (negative? indent) (set! indent 16))
			(out indent)
			(translate-exp bool? (+ 2 indent) sexp))
		      sense))
    (else
     (if (not sense)
	 (let ((tmp consequent))
	   (set! consequent alternate)
	   (set! alternate tmp)))
     (out indent "fif(")
     (translate-exp #t (+ 2 indent) test)
     (for-each (lambda (x)
		 (out CONTLINE #\, #\space)
		 (translate-exp bool? (+ 2 indent) x))
	       (list consequent alternate))
     (out CONTLINE ")"))))

(define (translate-ifdef test consequent alternate compile sense)
  (define tst
    (cond ((and (eq? 'PROVIDED? (car test))
		(pair? (cadr test))
		(eq? 'QUOTE (caadr test)))
	   (cadadr test))
	  ((string? (cadr test))
	   (cadr test))
	  ((symbol? (cadr test))
	   (cadr test))
	  (else (report "argument not identifier" test) test)))
  (report "translate-ifdef" test consequent alternate compile sense)
  (if (not sense)
      (let ((tmp consequent))
	(set! consequent alternate)
	(set! alternate tmp)))
  (if (provided? tst)
      (compile consequent)
      (compile alternate)))

(define (translate-case bool? indent cv clauses)
  (cond ((null? clauses) (translate-exp bool? indent (if bool? #f 0)))
	((and bool? (or (not (assq 'ELSE clauses))
			(not (car (last-pair (assq 'ELSE clauses))))))
	 (translate-exp
	  bool? indent
	  (cons
	   'or (map (lambda (clause)
		      `(and
			,(cond ((eq? 'ELSE (car clause))
				(cond
				 ((car (last-pair clause))
				  (report "else not #f?" clause)))
				#t)
			       ((= 1 (length (car clause)))
				`(= ,(case-eval (caar clause) clause) ,cv))
			       (else
				`(or ,@(map (lambda (v)
					      `(= ,(case-eval v clause) ,cv))
					    (car clause)))))
			,@(if (and
			       (pair? (cadr clause))
			       (eq? 'and (caadr clause)))
			      (cdadr clause)
			      (list (cadr clause)))))
		    clauses))))
	(else
	 (let ((clause (car clauses)))
	   (translate-exp
	    bool? indent
	    (if (eq? 'ELSE (car clause))
		(car (last-pair clause))
		`(if ,(cond ((= 1 (length (car clause)))
			     `(= ,(case-eval (caar clause) clause) ,cv))
			    (else
			     `(or ,@(map (lambda (v)
					   `(= ,(case-eval v clause) ,cv))
					 (car clause)))))
		     ,(car (last-pair clause))
		     (case ,cv ,@(cdr clauses)))))))))

(define (ram-ref block array idxs)
  ;;(print 'ram-ref block array idxs)
  (cond ((ram-table:row-retrieve block array) =>
	 (lambda (ram-row)
	   (print 'ram-row '= ram-row)
	   (destructuring-bind
	    (block name output write-enable address data)
	    ram-row
	    (if (and (eq? (car idxs) address)
		     (null? (cdr idxs)))
		(out CONTLINE output)
		(report "Bad use of ram array:" array idxs)))))
	(else #f)))

(define (translate-logaref indent array idxs)
  (let* ((bounds (vector-bounds array))
	 (name (if bounds (car bounds) array)))
    (cond ((not (ram-ref *block* array idxs))
	   (out CONTLINE "To_StdLogicVector(" name "(")
	   (integerize indent (car idxs))
	   (out CONTLINE "))")))))

(define (optionally-parenthesize bool? indent x width)
  (define bounds (vector-bounds x))
  (let ((objwid (bounds->width bounds)))
    (define need-pad? (and width
			   (not (eqv? 0 width))
			   (not (eqv? 0 objwid))
			   (not (eqv? width objwid))))
    (cond ((or need-pad?
	       (and (pair? x)
		    (not (memq (car x) '(NOT BIT-EXTRACT BIT-FIELD LOGBIT)))))
	   (out CONTLINE #\()
	   (cond (need-pad?
		  (out CONTLINE
		       (object->string (make-string (- width objwid) #\0))
		       " & ")
		  (translate-exp bool? CONTLINE x))
		 (else
		  (translate-exp bool? CONTLINE x)))
	   (out CONTLINE #\)))
	  (else (translate-exp bool? indent x)))))

(define (translate-and bool? indent exps)
  (case (length exps)
    ((0) (translate-exp bool? indent #t))
    ((1) (translate-exp bool? indent (car exps)))
    (else (cond (bool? (infix-translate-exp #t " AND " indent exps))
		(else (translate-exp
		       #f indent `(if (&& ,@(butlast exps 1))
				      ,(car (last-pair exps)) 0)))))))

(define (infix-translate-list bool? indent exps)
  (out CONTLINE #\()
  (cond ((not (null? exps))
	 (translate-exp bool? indent (car exps))
	 (set! exps (cdr exps))))
  (for-each (lambda (x)
	      (out CONTLINE #\, #\space)
	      (translate-exp bool? indent x))
	    exps)
  (out CONTLINE #\)))

(define (infix-translate-arith op indent exps)
  (define break #f)
  (define widths (map bounds->width (map vector-bounds exps)))
  (let ((max-width (apply max widths)))
    (cond ((not (null? exps))
	   (optionally-parenthesize #f indent (car exps) max-width)
	   (set! exps (cdr exps))))
    (set! break
	  (if (some (lambda (x)
		      (and (pair? x)
			   (not (memq (car x)
				      '(NOT BIT-EXTRACT BIT-FIELD LOGBIT?)))))
		    exps)
	      indent
	      CONTLINE))
    (for-each (lambda (x)
		(out break op)
		(optionally-parenthesize
		 #f (+ (if (char? op) 1 (string-length op)) indent) x max-width))
	      exps)))

(define (infix-translate-exp bool? op indent exps)
  (define break #f)
  (cond ((not (null? exps))
	 (optionally-parenthesize bool? indent (car exps) #f)
	 (set! exps (cdr exps))))
  (set! break
	(if (some (lambda (x)
		    (and (pair? x)
			 (not (memq (car x)
				    '(NOT BIT-EXTRACT BIT-FIELD LOGBIT?)))))
		  exps)
	    indent
	    CONTLINE))
  (for-each (lambda (x)
	      (out break op)
	      (optionally-parenthesize
	       bool? (+ (if (char? op) 1 (string-length op)) indent) x #f))
	    exps))

(define (infix-translate-pairwise op indent exps)
  (case (length exps)
    ((0) (translate-exp #t indent #t))
    ((1)
     (report 'one 'argument 'to op)
     (translate-exp #t indent #t))
    ((2)
     (infix-translate-exp
      #f (string-append " " (symbol->string op) " ") indent exps))
    (else
     (do ((exps exps (cdr exps))
	  (clauses '() (cons `(,op ,(car exps) ,(cadr exps)) clauses)))
	 ((null? (cdr exps))
	  (translate-exp #t indent (cons 'AND (reverse clauses))))))))

;;;; Netlists

(define (translate-power-connections)
  (define (power? name)
    (or (memq name '(vcc gnd))
	(let* ((str (symbol->string name))
	       (n (string-length str)))
	  (and (> n 2)
	       (memv (string-ref str 0) '(#\+ #\-))
	       (char=? (string-ref str (- n 1)) #\v)))))

  (define (out-net r p-nets non-p-nets)
    (out 3 (if (memq 'gnd p-nets)
	       "pulldown "
	       "pullup ")
	 (symbol-append 'r r)
	 " (")
    (for-each (lambda (net)
		(out CONTLINE net ", "))
	      non-p-nets)
    (let loop ((p-nets p-nets))
      (cond ((null? (cdr p-nets))
	     (out CONTLINE (car p-nets) ");"))
	    (else
	     (out CONTLINE (car p-nets) ", ")
	     (loop (cdr p-nets))))))

  ;; Returns list of power or ground net names.
  (define (out-all-nets row)
    (apply
     (lambda (name index sheet u-pins r-pins)
       (for-each
	(lambda (r-net)
	  (let loop ((nets (cadr r-net))
		     (p-nets '())
		     (non-p-nets '()))
	    (cond ((null? nets)
		   (if (pair? p-nets)
		       (out-net (car r-net) p-nets non-p-nets)))
		  ((power? (car nets))
		   (loop (cdr nets)
			 (cons (car nets) p-nets)
			 non-p-nets))
		  (else
		   (loop (cdr nets)
			 p-nets
			 (cons (car nets) non-p-nets))))))
	(map resistor-nets:row-retrieve r-pins)))
     row))

  (define (print-net name bounds)
    ((netlist 'for-each-row-in-order)
     (lambda (row)
       (out-all-nets row))
     name #f))

  ((netlist-aggregate 'for-each-row-in-order)
   (lambda (row)
     (apply print-net row)))

  ((netlist-non-aggregate 'for-each-row-in-order)
   (lambda (row)
     (print-net (car row) #f))))

(define (translate-u-nets)
  (let ((nets '()))
    ((netlist 'for-each-row-in-order)
     (lambda (row)
       (apply (lambda (name index sheet u-pins r-pins)
		(if (not sheet)
		    (set! nets (cons (list name) nets))))
	      row)))
    (out-decls "wire" nets))
  (letrec ((netlist-inverse-for-each (netlist-inverse 'for-each-row-in-order))
	   (ptag-for-each
	    (lambda (ptag proc)
	      (netlist-inverse-for-each
	       (lambda (row) (apply proc row)) ptag))))
    ((ptag-table 'for-each-row-in-order)
     (lambda (row)
       (apply (lambda (ptag pld-type signature optcom)
		(let ((module #f)
		      (instance #f))
		  (sscanf (symbol->string ptag) "%[^-]-%[^-]" module instance)
		  (cond (signature
			 (set! instance module)
			 (set! module (symbol-append *design* '- module))))
		  (cond (instance
			 (out 3 module " " (or instance module) "(")
			 (let ((count 0))
			   (ptag-for-each
			    ptag
			    (lambda (ptag pin net-name index)
			      (if (> count 0) (out CONTLINE ", "))
			      (let ((col (if (= 4 (remainder count 5))
					     7
					     CONTLINE)))
				(set! count (+ 1 count))
				(cond (index
				       (out col #\. pin
					    "(" net-name "(" index "))"))
				      (else
				       (out col #\. pin "(" net-name ")")))))))
			 (out CONTLINE ");")
			 (out 0)))))
	      row)))))

;; (db:read-u-ptag "cz-umap.scm")
;; (db:read-net "cz")
(define (translate-netlist)
  (translate-u-nets)
  (translate-power-connections))
;;(print 'doing-if) (trace translate-assignment translate-top) (set! *qp-width* 333)
;;(trace-all "/home/jaffer/synch/scm2vhdl.scm") (set! *qp-width* 333) (untrace translate-name lookup registered-input? numbers-in-expression? case-able? infix-translate-exp merge-bounds get-bounds) ; translate-exp
