;;;; scm2vrlg.scm: Program for translating SIMSYNCH code to VERILOG -*-scheme-*-
;;; Copyright (C) 1991, 1992, 1993, 1995, 1996, 1997, 1998, 2000 Aubrey Jaffer.

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

(define number-lines? #f)
;;; #t for C code to be prefixed with commmented line numbers.

;;(require 'common-list-functions)
(require 'chapter-order)
(require 'string-search)
(require 'fluid-let)
(require 'printf)
;;(require 'debug) (set! *qp-width* 100) (define qreport qpn)

;;; REPORT an error or warning
(define report
  (lambda args
    (display *verilog-output-name*)
    (cond (*current-signal-name*
	   (display ": In function `")
	   (display *current-signal-name*)
	   (display "': "))
	  (else (display ": ")))
    (display *verilog-output-line-number*)
    (display ": warning: ")
    (apply qreport args)))
(define qreport
  (lambda args
    (for-each (lambda (x) (write x) (display #\space)) args)
    (newline)))

;;; This allows us to test without generating files
(define *verilog-output* (current-output-port))
(define *verilog-output-name* "?")
(define *acf-output* (current-output-port))

;;; *index* is the bit of an integer -- not a subscript.
(define *index* #f)			;dynamic variable for looping vects.
(define *range* #f)			;dynamic variable for numeric constants.
(define *current-signal-name* #f)
(define *verilog-output-line-number* 0)
(define CONTLINE -80)

(define (vprintf . args)
  (apply fprintf *verilog-output* args)
  (newline *verilog-output*)
  (set! *verilog-output-line-number* (+ 1 *verilog-output-line-number*)))

(define (invertify sexp)
  (cond ((and (pair? sexp) (eq? 'NOT (car sexp)))
	 (cadr sexp))
	((boolean? sexp) (not sexp))
	(else (list 'NOT sexp))))
(define (negate pol sexp)
  (if pol (invertify sexp) sexp))

(define (indent->string indent)
  (string-append (make-string (quotient indent 8) #\tab)
		 (make-string (modulo indent 8) #\space)))

;;; OUT indents and displays the arguments
(define (out indent . args)
  (cond ((>= indent 0)
	 (newline *verilog-output*)
	 (set! *verilog-output-line-number* (+ 1 *verilog-output-line-number*))
	 (cond ((and number-lines? (> indent 0))
		(display "/*" *verilog-output*)
		(display *verilog-output-line-number* *verilog-output*)
		(display "*/" *verilog-output*)))
	 (display (indent->string indent) *verilog-output*)))
  (for-each (lambda (a)
	      (cond ((eq? '? a) (display "?" *verilog-output*))
		    ((symbol? a)
		     (translate-name a *verilog-output*))
		    (else (display a *verilog-output*))))
	    args))

(define (translate-name prname port)
  (define visible? #f)
  (define bounds-spec? #f)
  (define pin-rec ((pin-basename-table 'row:retrieve*)
		   (lambda (ptag)
		     (case ptag
		       ((int) #f)
		       (else #t)))
		   (cond ((vector-bounds prname) => car)
			 (else prname))))
  (define nstr (symbol->string prname))
  (case (length pin-rec)
    ((0) #f)
    ((1)
     (let ((ptag (caar pin-rec))
	   (pins (caddar pin-rec)))
       (set! pin-rec (pin-table:row-retrieve ptag (car pins))))
     (set! nstr
	   (case (caddr pin-rec)
	     ((internal)
	      (set! pin-rec ((signal-table 'get* 'defining-ptag) #f (cadr pin-rec)))
	      (case (length pin-rec)
		((0) #f)
		((1)
		 (display (car pin-rec) port)
		 (display #\_ port))
		(else (report "Multiple defining-ptags for name" prname pin-rec)))
	      nstr)
	     (else (string-upcase nstr)))))
    (else (report "Multiple pins for name" pin-rec) #f))
  (for-each
   (lambda (c)
     (let ((tc (cond ((char-alphabetic? c) c)
		     ((char-numeric? c) c)
		     ((char=? c #\.)
		      (cond (bounds-spec? (set! bounds-spec? #f) #\:)
			    (else #f)))
		     ((char=? c #\[) (set! bounds-spec? #t) #\[)
		     ((char=? c #\]) (set! bounds-spec? #f) #\])
		     ((char=? c #\-) #\_)
		     ((char=? c #\_) #\_)
		     ((char=? c #\?) "_P")
		     ((char=? c #\%) "Percent")
		     ((char=? c #\:) (cond (bounds-spec? #\:)
					   (else #\_)))
		     (else #f))))
       (cond (tc (set! visible? #t)
		 (display tc port)))))
   (string->list nstr))
  (case (string-ref nstr (+ -1 (string-length nstr)))
    ((#\_ #\-) (display #\n
			;;(if (char-upper-case? (string-ref nstr 0)) #\N #\n)
			port)))
  (if (not visible?) (report "C-invisible symbol?" prname)))

(define (outcomment indent str)
  (out indent "//"
       (string-subst str "\n" (string-append "\n" (indent->string indent)
					     "//"))))

(define (vrlg:revectorize sym ind1 ind2)
  (revectorize ":" sym ind1 ind2))

(define (descmfilify file)
  (let ((sl (string-length file)))
    (cond ((< sl 4) file)
	  ((string-ci=? (substring file (- sl 4) sl) ".scm")
	   (substring file 0 (- sl 4)))
	  (else file))))

;;; TRANSLATE ptags in design.
(define (translate-design design)
  (for-each-ptag
   (lambda (tag pld-type signature optcom)
     (printf "translating %s for ptag %s\n" design tag)
     (synch:check-flags '(SYNCHRONOUS-RESET NO-RTL-SIMULATION))
     (translate-tag design tag))))

(define (call-with-verilog-output design module suffix proc)
  (define ofile
    (string-append (symbol->string design) "_" (symbol->string module)
		   ;;(number->string time-stamp:revision)
		   suffix))
  (printf "generating %s\n" ofile)
  (fluid-let ((*verilog-output* (open-file ofile "w"))
	      (*verilog-output-name* ofile)
	      (*verilog-output-line-number* 0))
    (proc)
    (close-output-port *verilog-output*)))

(define (numbers-in-expression? tree)
  (let walk ((tree tree))
    (cond ((number? tree))
	  ((not (pair? tree)) #f)
	  ((not (list? tree)) #f)
	  ((memq (car tree) '(logbit? =)) #f) ; ash bit-extract
	  ((some number? (cdr tree)))
	  (else (some walk (cdr tree))))))

(define (registered-input? name)
  (and (symbol? name)
       (let ((name (symbol->string name)))
	 (and (> (string-length name) 4)
	      (string=? "pre:" (substring name 0 4))
	      (string->symbol (substring name 4 (string-length name)))))))

(define (generate-verilog-header design tag)
  (require 'posix-time)
  (let* ((ptag (or (ptag->pld tag) "PALCE630"))
	 (signature (ptag->sig tag)))
    (vprintf "")
    (vprintf "//TITLE     %s" (symbol-append design '- tag))
    (vprintf "//FILE      %s" *verilog-output-name*)
    (vprintf "//PATTERN   %s" (or time-stamp:configuration ""))
    (vprintf "//REVISION  %a" time-stamp:revision)
    (and time-stamp:author
	 (vprintf "//AUTHOR    %s" time-stamp:author))
    (and time-stamp:company
	 (vprintf "//COMPANY   %s" time-stamp:company))
    (vprintf "//DATE      %s" (ctime (or time-stamp:time (current-time))))
    (vprintf "//CHIP  _%s  %s" tag ptag)
    (cond ((and signature
		(> (string-length ptag) 4)
		(string-ci=? (substring ptag 0 4) "mach"))
	   (if (not (eqv? signature (logand #xffffffff signature)))
	       (report "signature value out of range"
		       (number->string signature 16))
	       (vprintf "//SIGNATURE #h%x" signature))))))

(define (out-decl type args coms)
  (define (out-line nm bnds fbnds)
    (out 3 type)
    (if (and (pair? bnds) (car bnds))
	(out CONTLINE " [" (car bnds)
	     #\: (cadr bnds) #\]))
    (out CONTLINE #\space nm)
    (cond ((and (pair? bnds) (pair? (cddr bnds)))
	   (out CONTLINE " [")
	   (translate-exp #f 5 (cadddr bnds))
	   (out CONTLINE #\:)
	   (translate-exp #f 5 (car (last-pair fbnds)))
	   (out CONTLINE #\]))))

  (if (pair? args)
      (let loop ((args args)
		 (iarg 1)
		 (last-bnds #f)
		 (coms (or coms (map (lambda (arg) #f) args))))
	(if (null? args)
	    (out CONTLINE #\;)
	    (apply
	     (lambda (nm . fbnds)
	       (define bnds (butlast fbnds 1))
	       (cond ((car coms)
		      (if last-bnds (out CONTLINE #\;))
		      (out 0)
		      (outcomment 10 (car coms))
		      (out-line nm bnds fbnds)
		      (loop (cdr args) 1 'no-match (cdr coms)))
		     ((equal? bnds last-bnds)
		      (out CONTLINE #\,)
		      (cond ((zero? (modulo iarg 6))
			     (out 10 nm))
			    (else (out CONTLINE #\space nm)))
		      (loop (cdr args) (+ 1 iarg) bnds (cdr coms)))
		     (else
		      (if last-bnds (out CONTLINE #\;))
		      (out-line nm bnds fbnds)
		      (loop (cdr args) 1 bnds (cdr coms)))))
	     (car args))))))

(define (translate-constant-definition var def optcom)
  (set! *current-signal-name* var)
  (cond (optcom (out 0) (outcomment 0 optcom)))
  (out 3 "parameter ")
  (translate-name var *verilog-output*) ;name
  (out CONTLINE " = ")
  (cond ((boolean? def)
	 (out CONTLINE (if def #\1 #\0)))
	((and (number? def) (negative? def))
	 (out CONTLINE "1'bx"))
	(else (translate-exp #f 8 def)))
  (out CONTLINE ";")
  (set! *current-signal-name* #f))

(define (generate-module-header design tag block)
  ;;(print 'generate-module-header design tag block)

  (call-with-classified-pins
   tag block
   (lambda (outputs inouts inputs registers)
     (define block-controls ((block-table 'row:retrieve) block))
     (cond
      (block-controls
       ;;(print 'module block 'clock= (cadr block-controls) 'reset= (caddr block-controls))
       (set! inputs (merge-bounds (list (cadr block-controls)) inputs))
       ;;(set! inputs (merge-bounds (list (caddr block-controls)) inputs))
       (fprintf *spew-output* ";; State dump for %s %s %s\n"
		design tag block)
       (pretty-print
	`(dump:verilog
	  ,(string-append (symbol->string design) "-spew.dat")
	  ,@(map (lambda (sig)
		   (case (length sig)
		     ((1) (car sig))
		     ((4)
		      (set! sig (cadddr sig))
		      (case (length sig)
			((1) (car sig))
			(else
			 `(logior
			   ,@(map (lambda (sig)
				    `(ash ,sig ,(caddr (vector-bounds sig))))
				  (reverse
				   (map string->symbol
					(sort (map symbol->string sig)
					      chap:string<?))))))))
		     (else (report 'funny-sig sig))))
		 (apply append (map reverse
				    (map remove-duplicates
					 (list inputs inouts outputs))))))
	*spew-output*)
       (newline *spew-output*)))
     (instantiate 'module
		  (if block (symbol-append design '_ block) design)
		  inputs inouts outputs)
     (cond (block
	    (out 0)
	    (outcomment 0 " Constants") (out 0)
	    ((constant-table 'for-each-row-in-order)
	     (lambda (constant-row)
	       (apply
		(lambda (name value optcom)
		  (cond ((and (not (used-in-ptag:get-ptag name tag))
			      (not (boolean? value))))
			((or (boolean? value) (number? value))
			 (translate-constant-definition name value optcom))))
		constant-row)))))
     (out 0)
     (out-decl "input" inputs (map signal->comment inputs))
     (out-decl "output" outputs (map signal->comment outputs))
     (cond ((not (provided? 'lift-bus))
	    (out 0)
	    (outcomment 0 "Bidirectional signals") (out 0)
	    (out-decl "inout" inouts (map signal->comment inouts))
	    (out-decl "tri" inouts #f)))
     (cond (block
	    (out 0)
	    (outcomment 0 "Registers") (out 0)
	    (out-decl "reg" registers (map signal->comment registers))))
     inouts)))

(define (generate-module-footer design tag block)
  (out 0)
  (out 0 "endmodule // " (if block (symbol-append design '_ block) design))
  (out 0))

(define (generate-standard-includes design tag)
  (vprintf "") (vprintf "")
  (vprintf "//  Verilog includes")
  (out 0 "`include " "\"logical.v\"")
  ;;(out 0 "`include " "\"asyn_ram_256x8.v\"")
  (out 0))

(define (generate-module-includes module-files)
  (vprintf "") (vprintf "")
  (vprintf "//  Verilog module includes")
  (for-each (lambda (inc)
	      (out 0 "`include " "\"" inc "\""))
	    module-files)
  (out 0))

(define (generate-inc-file design *tris* *inouts*)
  (define paireds '())
  (call-with-verilog-output
   design 'inc ".v"
   (lambda ()
     (out 0)
     (outcomment 0 "Bidirectional signals") (out 0)
     (out-decl "inout" *inouts* #f)
     (out-decl "tri" *inouts* #f)
     (for-each
      (lambda (row)
	(apply (lambda (paired-input output-enable name)
		 (set! paireds (merge-bounds (get-bounds name) paireds)))
	       row))
      *tris*)
     (out-decl "wire" paireds #f)
     (out 0)
     (for-each
      (lambda (row)
	(apply (lambda (paired-input output-enable name)
		 (define bnds (vector-bounds name))
		 (out 3 "assign " paired-input " = ")
		 (optionally-parenthesize (not bnds) 8 output-enable)
		 (out CONTLINE " ? " name " : "
		      (if bnds
			  (+ 1 (abs (apply - (cdr bnds))))
			  1)
		      "'bz;"))
	       row))
      *tris*)
     (out 0))))

(define-syntax debind:if
  (syntax-rules ()
    ((_ (?lit ...) ?pat ?obj ?conseq ?alt)
     (letrec-syntax ((debind1
		      (syntax-rules (?lit ...)
			((_ (??first . ??rest) ??obj ??conseq ??alt)
			 (if (pair? ??obj)
			     (let ((carv (car ??obj))
				   (cdrv (cdr ??obj)))
			       (debind1 ??first carv
					(debind1 ??rest cdrv ??conseq ??alt)
					??alt))
			     ??alt))
			((_ ?lit ??obj ??conseq ??alt)
			 (if (eq? '?lit ??obj) ??conseq ??alt))
			...
			((_ () ??obj ??conseq ??alt)
			 (if (null? ??obj) ??conseq ??alt))
			((_ ??name ??obj ??conseq ??alt)
			 (let ((??name ??obj)) ??conseq)))))
       (debind1 ?pat ?obj ?conseq ?alt)))))

(define-syntax destructuring-case
  (syntax-rules ()
    ((_ "RECUR" ?obj ?lits (?pat ?body1 ?body ...))
     (debind:if ?lits ?pat ?obj (begin ?body1 ?body ...) #f))
    ((_ "RECUR" ?obj ?lits (?pat ?body1 ?body ...) ?clause ...)
     (debind:if ?lits ?pat ?obj (begin ?body1 ?body ...)
		(destructuring-case "RECUR" ?obj ?lits ?clause ...)))
    ((_   ?obj (?lit ...)
	(?pat ?body1 ?body ...) ...)
     (let ((obj ?obj))
       (destructuring-case "RECUR" obj (?lit ...)
			   (?pat ?body1 ?body ...) ...)))))

(define (process-rams)
  (define (make-name base suffix bnds)
    (string->symbol
     (sprintf #f "%s-%s[%d:%d]"
	      (symbol->string base) (symbol->string suffix)
	      (car bnds) (cadr bnds))))
  (define (parse-ram-next ram-name sexp)
    (define (err . objs)
      (apply report "asyn-ram: unexpected form for next-function:"
	     ram-name sexp objs))
    (define (return name1 name2 write-enable data address)
      (if (eq? ram-name name1 name2)
	  (list write-enable address data)
	  (err)))
    (destructuring-case
     sexp
     (IF ARRAY-SET)
     ((IF test (ARRAY-SET name1 data address) name2)
      (return name1 name2 test data address))
     ((IF test name2 (ARRAY-SET name1 data address))
      (return name1 name2 `(not ,test) data address))
     (fail (err))))
  ((ram-table 'for-each-row-in-order)
   (lambda (ram-row)
     (destructuring-bind
	 (block name address data output write-enable)
	 ram-row
       (destructuring-bind
	   (block name ptag signal-type initial-state next-function order)
	   (signal-table:row-retrieve block name)
	 (fluid-let ((*block* block))
	   (let ((bnds (vector-bounds name)))
	     (cond ((and (pair? bnds)
			 (equal? '(7 0) (cdr bnds)))
		    (let ((output (make-name (car bnds) 'output (cdr bnds)))
			  (parsed (parse-ram-next name next-function)))
		      (synch:define-signal output ptag 'macro #f #f)
		      (ram-table:row-update
		       (append (list block name output) parsed)))
		    (let ((pin-row (pin-table:row-retrieve ptag name)))
		      (set-car! (cddr pin-row) 'library-module) ;pin-type
		      ((pin-table 'row:update) pin-row)))
		   (else
		    (report "unsupported asyn-ram width" block name))))))))))

(define (named-instantiate module name formals bools actuals)
  (out 3 module name "(")
  (letrec ((outarg (lambda (formal bool? actual)
		     (out 20 "." formal "(")
		     (translate-exp bool? CONTLINE actual)
		     (out CONTLINE ")"))))
    (outarg (car formals) (car bools) (car actuals))
    (for-each (lambda (formal bool? actual)
		(out CONTLINE ",")
		(outarg formal bool? actual))
	      (cdr formals)
	      (cdr bools)
	      (cdr actuals)))
  (out 20 ");"))

(define (instantiate-asyn-ram block name)
  (destructuring-bind
      (block name . rest)
      (ram-table:row-retrieve block name)
    (let ((bnds (vector-bounds name)))
      (named-instantiate
       "asyn_ram_256x8_iror " (car bnds)
       '("Q" "WE" "Address" "Data")
       '(#f  #t #f  #f)
       rest))))

;;; TRANSLATE-TAG - translate top level forms.
(define (translate-tag design tag)
  (define *lift-bus* (provided? 'lift-bus))
  (define *tris* '())
  (define *inouts* '())
  (define module-files '())
  (process-rams)
  ((block-table 'for-each-row-in-order)
   (lambda (block-row)
     (apply
      (lambda (block-name clock reset clock-frequency synchronous-reset? optcom)
	(call-with-verilog-output
	 design block-name ".v"
	 (lambda ()
	   (set! module-files (cons *verilog-output-name* module-files))
	   (if (file-exists? "Copyrite.v")
	       (call-with-input-file "Copyrite.v"
		 (lambda (inport)
		   (do ((line (read-line inport) (read-line inport)))
		       ((eof-object? line))
		     (out 0 line))))
	       (generate-verilog-header design tag))
	   ;; (out 4 "// Block " block-name)
	   (cond (optcom (out 0) (outcomment 0 optcom)))
	   (generate-module-header design tag block-name)
	   (generate-standard-includes design tag)
	   (let ((wire-lst '())
		 (com-lst '()))
	     (signal-table:for-each-row
	      (lambda (signal-row)
		(apply
		 (lambda (block name defining-ptag signal-type
				initial-state next-function order)
		   (define optcom (signal->comment name))
		   (and (or (not defining-ptag) (eq? tag defining-ptag))
			(case signal-type
			  ((macro)
			   (set! wire-lst (cons (get-bounds name) wire-lst))
			   (set! com-lst (cons optcom com-lst))))))
		 signal-row))
	      block-name)
	     (out-decl "wire" wire-lst com-lst))
	   (signal-table:for-each-row
	    (lambda (signal-row)
	      (apply
	       (lambda (block name defining-ptag signal-type
			      initial-state next-function order)
		 (define optcom (signal->comment name))
		 ;;(print 'fluid-let '*block* *block* 'block block signal-row)
		 (fluid-let ((*block* block))
		   (set! initial-state (ceval initial-state))
		   (and
		    (or (not defining-ptag) (eq? tag defining-ptag))
		    (or (pin-table:row-retrieve tag name)
			(eq? signal-type 'macro))
		    (cond (optcom (out 0) (outcomment 0 optcom)))
		    (case signal-type
		      ((macro)
		       (cond ((or initial-state next-function)
			      (out 3 "assign " name " = ")
			      ;;(print 'name= name next-function)
			      (translate-exp (not (vector-bounds name))
					     8 next-function)
			      (out CONTLINE ";"))))
		      ((registered registered-input combinatorial)
		       (let ((bnds (vector-bounds name))
			     (flags (synch:signal-flags block-name name)))
			 (and
			  (not (string-index
				(symbol->string (if bnds (car bnds) name)) #\:))
			  (cond ((eqv? *xxxx* initial-state))
				((and (boolean? initial-state) bnds)
				 (report
				  name 'vector 'initialized initial-state '?))
				((and (number? initial-state) (not bnds))
				 (report
				  name 'vector? 'initialized initial-state))))
			 (cond ((memq 'NO-RTL-SIMULATION flags)
				(out 0 "`ifdef SIMULATE")
				(out 0 "`else")))
			 (apply
			  (lambda (pin-type pin-number output-enable paired-input defining-block)
			    (cond ((and (eq? paired-input name)
					(not (eq? #t output-enable)))
				   (set! paired-input
					 (symbol-append 'pin- paired-input))))
			    (cond ((and paired-input
					(not (eq? paired-input #t))
					(not (eq? output-enable #t)))
				   ;;(print 'assign paired-input output-enable name)
				   (set! *tris* (cons (list paired-input
							    output-enable
							    name)
						      *tris*))
				   (cond
				    ((not *lift-bus*)
				     (out 3 "assign " paired-input " = ")
				     (optionally-parenthesize
				      (not bnds) 8 output-enable)
				     (out CONTLINE " ? " name " : "
					  (if bnds
					      (+ 1 (abs (apply - (cdr bnds))))
					      1)
					  "'bz;")))))
			    (cond ((not (eq? paired-input #t))
				   (cond
				    ((or
				      (and (number? initial-state)
					   (not (eqv? *xxxx* initial-state)))
				      (boolean? initial-state)
				      (and (pair? initial-state)
					   (not (memq (car initial-state)
						      '(MAKE-RAM MAKE-RAM-ARRAY
								 MAKE-ASYN-RAM MAKE-FIFO)))))
				     (out 3 "always @ (posedge " clock)
				     (if (not (or synchronous-reset?
						  (memq 'synchronous-reset flags)))
					 (out CONTLINE " or negedge " reset))
				     (out CONTLINE ") begin")
				     (translate-top
				      `(define ,name (if (not ,reset)
							 ,initial-state
							 ,next-function)))
				     (out 3 "end //always"))
				    ((and (pair? initial-state)
					  (eq? (car initial-state) 'MAKE-ASYN-RAM))
				     (instantiate-asyn-ram block name))
				    ((or (eqv? *xxxx* initial-state)
					 (equal? name initial-state)
					 (equal? name reset))
				     (out 3 "always @ (posedge " clock ") begin")
				     ;;(qreport 'non-reset 'register: name)
				     (translate-top
				      `(define ,name ,next-function))
				     (out 3 "end //always"))
				    (else
				     (report "funny initial value for" name ':
					     initial-state)
				     (out 3 "always @ (posedge " clock ") begin")
				     ;;" or " " negedge " reset
				     ;;(if (not ,reset) #t ...)
				     (translate-top
				      `(define ,name ,next-function))
				     (out 3 "end //always")))))
			    (out 3))
			  (cddr (pin-table:row-retrieve tag name)))
			 (cond ((memq 'NO-RTL-SIMULATION flags)
				(out 0 "`endif //not SIMULATE")))))))))
	       signal-row))
	    block-name)
	   ;;(out 4 " // End Block " block-name)
	   (generate-module-footer design tag block-name))))
      block-row)))
  (set! *inouts* (generate-stimulus-file design tag module-files))
  (if *lift-bus* (generate-inc-file design *tris* *inouts*)))

(define (generate-stimulus-file design tag module-files)
  (define *inouts* #f)
  (call-with-verilog-output
   design 'stim ".v"
   (lambda ()
     (generate-verilog-header design tag)
     (set! *inouts* (generate-module-header design tag #f))
     ((block-table 'for-each-row-in-order)
      (lambda (block-row)
	(apply
	 (lambda
	     (block-name clock reset clock-frequency synchronous-reset? optcom)
	   (call-with-classified-pins
	    tag block-name
	    (lambda (outputs inouts inputs registers)
	      (cond (optcom (out 0) (outcomment 0 optcom)))
	      (instantiate (symbol-append design '_ block-name)
			   (symbol-append tag '_ block-name)
			   inputs inouts outputs))))
	 block-row)))
     (generate-module-footer design tag #f)
     (generate-module-includes module-files)
     (generate-stimulus design tag)))
  *inouts*)

(define (instantiate design tag inputs inouts outputs)
  (define loop (lambda (i args)
		 (cond ((pair? args)
			(let ((a (car args)))
			  (cond ((eqv? 'module design)
				 (out (if (zero? (modulo i 6)) 6 CONTLINE))
				 (translate-name (car a) *verilog-output*))
				(else
				 (out (if (zero? (modulo i 3)) 6 CONTLINE) #\.)
				 (translate-name (car a) *verilog-output*)
				 (out CONTLINE #\()
				 (translate-name (car a) *verilog-output*)
				 (out CONTLINE #\))))
			  (if (pair? (cdr args))
			      (out CONTLINE ", "))
			  (loop (+ i 1) (cdr args)))))))
  (define outsigs (lambda (name sigs)
		    (cond ((not (null? sigs))
			   (out 12 "//module " name)
			   (loop 0 sigs)))))
  (out 0)
  (out (if (eqv? 'module design) 0 3) design " " tag "(")
  (outsigs "inputs" inputs)
  (cond ((not (provided? 'lift-bus))
	 (if (pair? inouts) (out CONTLINE ","))
	 (outsigs "inouts" inouts)))
  (if (pair? outputs) (out CONTLINE ","))
  (outsigs "outputs" outputs)
  (out CONTLINE ");"))

(define (generate-stimulus design tag)
  (call-with-classified-pins
   tag #f
   (lambda (outputs inouts inputs registers)
     (out 0 "`ifdef exemplar")
     (out 0 "// No stimulus module")
     (out 0 "`else")
     (out 0 "module " 'stimulus ";")
     (out 0)
     (outcomment 0 "module inputs") (out 0)
     (out-decl "reg" inputs #f)
     (out 0)
     (outcomment 0 "module outputs") (out 0)
     (out-decl "wire" outputs #f)
     (out 0)
     (outcomment 0 "module bidirectional signals") (out 0)
     (out-decl "wire" inouts #F)
     (instantiate design tag inputs inouts outputs)
     (out 0)
     (out 0 'initial)
     (out 0 'begin)
     (for-each
      (lambda (input)
	(translate-assignment
	 (if (null? (cdr input)) (car input) (caar (last-pair input)))
	 3 #f 0))
      inputs)
     (out 0 'end)
     (out 0 "endmodule // " 'stimulus)
     (out 0 "`endif")
     (out 0))))

(define (for-each-index proc bounds)
  (let* ((first (cadr bounds))
	 (last (caddr bounds))
	 (range (- last first)))
    (if (positive? range)
	(do ((i 0 (+ 1 i))
	     (j first (+ 1 j)))
	    ((> i range))
	  (fluid-let ((*index* i))
	    ;;(print '+range i j)
	    (proc j)))
	(do ((i (- range) (+ -1 i))
	     (j first (+ -1 j)))
	    ((negative? i))
	  (fluid-let ((*index* i))
	    ;;(print '-range i j)
	    (proc j))))))

(define (for-each-value proc bounds)
  (let* ((first (cadr bounds))
	 (last (caddr bounds))
	 (range (- last first)))
    (if (not (zero? (min first last)))
	(report "mapping over non-consecutive values" bounds))
    (set! range (ash 2 (abs range)))
    (do ((i 0 (+ 1 i)))
	((>= i range))
      (fluid-let ((*index* i))
	(proc i)))))

;;; TRANSLATE-TOP - translate top level form sexp.
(define (translate-top sexp)
  (cond ((symbol? sexp) (set! *current-signal-name* sexp))
	((and (pair? sexp) (eq? (car sexp) 'QUOTE))
	 (set! *current-signal-name* (cadr sexp)))
	((string? sexp) (outcomment 0 sexp) (out 0))
	((not (pair? sexp))
	 (report "top level atom?" sexp))
	(else (case (car sexp)
		((define) (translate-assignment (cadr sexp) 3 #f (caddr sexp)))
		(else (report "statement not in procedure" sexp)))))
  (out 0)
  (set! *current-signal-name* #f))

(define (ifable? rhs)
  (and (pair? rhs)
       (case (car rhs)
	 ((CASE CASEV IF AND NOT ARRAY-SET LOGASET) #t) ; NUMBER-OR OR
	 ((SYNCH:PRE) (ifable? (signal-table:get-next-fun *block* (cadr rhs))))
	 (else #f))))

(define (translate-assignment lhs indent pol rhs)
  (define bounds (vector-bounds lhs))
  (set! *current-signal-name* lhs)
  (cond ((ifable? rhs)
	 (case (car rhs)
	   ((NOT)
	    (translate-assignment lhs indent (not pol) (cadr rhs)))
	   ((SYNCH:PRE)
	    (let ((expn (signal-table:get-next-fun *block* (cadr rhs))))
	      (if (not expn) (report "returned #f" rhs))
	      (translate-exp (not bounds) indent expn)))
	   ((CASE CASEV) (out indent "case (")
	    (translate-exp #f (+ 6 indent) (cadr rhs))
	    (out CONTLINE ")")
	    (for-each (lambda (clause)
			(translate-case-clause lhs (+ 2 indent) pol clause))
		      (cddr rhs))
	    (out indent "endcase"))
	   ((IF)
	    (translate-assign-if lhs indent pol rhs #t))
	   ((AND) (translate-assignment
		   lhs
		   indent
		   pol
		   (case (length rhs)
		     ((1) #t)
		     ((2) (cadr rhs))
		     (else `(if ,(butlast rhs 1) ,(car (last rhs 1)) #f)))))
	   ((ARRAY-SET LOGASET)
	    (if (not (equal? lhs (cadr rhs)))
		(report 'array-set lhs rhs))
	    (let ((name (if bounds (car bounds) lhs)))
	      (out indent)
	      (translate-name name *verilog-output*)
	      (out CONTLINE #\[)
	      (translate-exp #f (+ 10 indent) (cadddr rhs))
	      (out CONTLINE #\])
	      (out (+ 4 indent) " <= ")	;name
	      (translate-exp #f (+ 8 indent) (negate pol (caddr rhs)))
	      (out CONTLINE #\;)))
	   (else (report "ifable expression not translated" rhs))))
	((and (equal? lhs rhs) (not pol)
	      (not (null? (array-bounds *block* lhs))))
	 (out CONTLINE " ;"))
	(else
	 (out indent)
	 (translate-name lhs *verilog-output*)
	 (out CONTLINE " <= ")		;name
	 (translate-exp (not bounds) (+ 8 indent) (negate pol rhs))
	 (out CONTLINE #\;))))

(define (case-eval dat clause)
  (cond ((and (pair? dat) (memq (car dat) '(UNQUOTE-SPLICING UNQUOTE)))
	 (if (not (constant-table:row-retrieve (cadr dat)))
	     (report (cadr dat) "case datum not constant in " clause))
	 (cadr dat))
	(else dat)))

(define (translate-case-clause lhs indent pol clause)
  (cond ((eq? 'ELSE (car clause))
	 (let ((lc (car (last-pair clause))))
	   (cond ((and (pair? lc)
		       (memq (car lc) '(synch:chat synch:warn synch:error))))
		 ((equal? lhs lc))
		 (else
		  #+f
		  (if (not (member lc '(0 #f)))
		      (report "case else not (0) #f?" lc))
		  (out indent "default :")
		  (translate-assignment lhs (+ 2 indent) pol lc)))))
	((= 1 (length (car clause)))
	 (let ((dat (caar clause)))
	   (out indent (case-eval dat clause) " :"))
	 (translate-assignment lhs (+ 2 indent)
			       pol (car (last-pair clause))))
	(else
	 (for-each (lambda (match)
		     (translate-case-clause
		      lhs indent pol (cons (list match) (cdr clause))))
		   (car clause)))))

(define (translate-assign-if lhs indent pol rhs sense)
  (define test (cadr rhs))
  (define consequent (caddr rhs))
  (define alternate (cadddr rhs))
  (case (and (pair? test) (car test))
    ((NOT) (translate-assign-if lhs indent pol
				`(if ,(cadr test) ,consequent ,alternate)
				(not sense)))
    ((DEFINED? PROVIDED?)
     (translate-ifdef test consequent alternate
		      (lambda (sexp)
			(if (negative? indent) (set! indent 16))
			(out indent)
			(translate-assignment lhs (+ 2 indent) pol sexp))
		      sense))
    (else
     (out indent (string-append "if " "(" (if sense "" "!")))
     (translate-exp #t (+ 2 indent) test)
     (out CONTLINE ")")
     (translate-assignment lhs (+ 2 indent) pol consequent)
     (cond ((null? (cdddr rhs)) (report 'dangling-if? lhs rhs))
	   ;;((equal? lhs alternate))
	   (else
	    (out indent "else")
	    (translate-assignment lhs (+ 2 indent) pol alternate))))))

;;; LOOKUP - translate from table or return arg as string
(define (lookup arg tab)
  (let* ((p (assq arg tab))
	 (l (if p (cdr p) arg)))
    (if (symbol? l) (symbol->string l) l)))

(define (out-uint sexp)
  (define len (integer-length sexp))
  (if (>= len 32)
      (out CONTLINE
	   32
	   "'h"
	   (if (<= 0 sexp 15) "0" "")
	   (number->string sexp 16))
      (out CONTLINE
	   (if (<= 0 sexp 9) "0" "")
	   (number->string sexp))))

;;; TRANSLATE-EXP - translate expression
(define (translate-exp bool? indent sexp)
  (cond ((not (pair? sexp))		;atoms
	 (cond ((string? sexp)
		(out CONTLINE #\" sexp #\"))
	       ((boolean? sexp)
		(out CONTLINE (case sexp ((#f) "0") ((#t) "1"))))
	       ((and *index* (symbol? sexp) (vector-bounds sexp))
		(translate-exp bool? indent (if (negative? *index*)
						#f
						`(logbit? ,*index* ,sexp))))
	       ((and *index* (integer? sexp))
		(out CONTLINE (if (and (not (negative? *index*))
				       (logbit? *index* sexp))
				  "1" "0")))
	       ((symbol? sexp) (out CONTLINE sexp))
	       ((number? sexp)
		(if *range*
		    (let* ((num (if (positive? *range*)
				    (reverse-bit-field sexp 0 *range*)
				    sexp))
			   (str (number->string num 2)))
		      (print ">>>> Time to debug numbers within *range*.")
		      (out CONTLINE
			   (number->string (abs *range*))
			   "'b"
			   (make-string (- (abs *range*) (string-length str))
					#\0)
			   str))
		    (cond ((zero? sexp) (out CONTLINE "00"))
			  ((negative? sexp)
			   (report "negative number:" sexp)
			   (set! sexp (abs sexp))
			   (out CONTLINE " - "
				(if (<= 0 sexp 9) "0" "")
				(number->string sexp)))
			  (else
			   (out-uint sexp)))))
	       (else (report "unexpected datatype:" sexp)
		     (out CONTLINE sexp))))
	(else
	 (case (car sexp)
	   ;; These vectorized functions are for multiple bit signals:
	   ((< <= > >=)
	    (cond ((not (= 2 (length (cdr sexp))))
		   (report "= should take two arguments" sexp)))
	    (translate-<> indent (car sexp) (cadr sexp) (caddr sexp)))
	   ((= eqv?)
	    (cond ((not (= 2 (length (cdr sexp))))
		   (report "= should take two arguments" sexp)))
	    (translate-= indent (cadr sexp) (caddr sexp)))
	   ((ZERO?)
	    (out CONTLINE "00 == (")
	    (translate-exp #t indent (cadr sexp))
	    (out CONTLINE ")"))
	   ((LOGNOT)
	    (cond ((symbol? (cadr sexp))
		   (out CONTLINE "~")
		   (translate-exp bool? (+ 2 indent) (cadr sexp)))
		  (else
		   (out CONTLINE "~" "(")
		   (translate-exp bool? (+ 2 indent) (cadr sexp))
		   (out CONTLINE ")"))))
	   ((MODULO)
	    (translate-exp bool? (+ 2 indent)
			   `(logand (- ,(caddr sexp) 1)
				    ,(cadr sexp))))
	   ((LOGIOR LOGAND LOGXOR - / QUOTIENT) ; MODULO
	    (infix-translate-exp
	     #f
	     (lookup (car sexp)
		     '((LOGIOR . " | ") (LOGAND . " & ")
		       (LOGXOR . " ^ ") (- . " - ") (/ . " / ")
		       (MODULO . " % ") (QUOTIENT . " / ")))
	     indent
	     (cdr sexp)))
	   ((+)
	    (infix-translate-arith " + " (+ 5 indent) (cdr sexp)))
	   ((*)
	    (infix-translate-arith " * " (+ 5 indent) (cdr sexp)))
	   ((LOGTEST)
	    (translate-exp #t indent
			   `(not (zero? (logand ,@(cdr sexp))))))
	   ((NUMBER-OR) (infix-translate-exp #f " | " indent (cddr sexp)))
	   ((ASH ARITHMETIC-SHIFT)
	    (cond (*index* (fluid-let ((*index* (- *index* (caddr sexp))))
			     (translate-exp bool? indent (cadr sexp))))
		  ((not (number? (caddr sexp)))
		   (report "ash not handled" sexp)
		   (infix-translate-exp #f " << " indent (cdr sexp)))
		  ((negative? (caddr sexp))
		   (infix-translate-exp #f " >> " indent
					(list (cadr sexp) (- (caddr sexp)))))
		  (else
		   (out CONTLINE "{(")
		   (translate-exp #t (+ indent 5) (cadr sexp))
		   (out CONTLINE ")," (caddr sexp) "'b0}"))))
	   ((BIT-EXTRACT BIT-FIELD)
	    ;;(print (car sexp) (cdr sexp))
	    (let* ((bounds (vector-bounds (cadr sexp)))
		   (lowbnd (and bounds (min (cadr bounds) (caddr bounds)))))
	      (cond ((not bounds)
		     (report "bit extract of non-symbol?" sexp))
		    ((and (number? (caddr sexp))
			  (number? (cadddr sexp)))
		     (translate-name (vrlg:revectorize
				      (cadr sexp)
				      (+ lowbnd (caddr sexp))
				      (+ lowbnd (cadddr sexp) -1))
				     *verilog-output*))
		    (else
		     (out CONTLINE 'bit-field)
		     (infix-translate-list #f (+ 2 indent) (cdr sexp))))))
	   ((BOOLEANS-TO-NUMBER)
	    (out CONTLINE "{")
	    (translate-exp #t (+ indent 5) (cadr sexp))
	    (for-each (lambda (bit)
			(out CONTLINE ",")
			(out (+ indent 5))
			(translate-exp #t (+ indent 5) bit))
		      (cddr sexp))
	    (out CONTLINE "}"))
	   ((LOGBIT?)
	    (cond
	     ((and (number? (cadr sexp)) (negative? (cadr sexp))
		   (translate-exp bool? indent #f)))
	     ((and (number? (cadr sexp)) (symbol? (caddr sexp)))
	      (let* ((bounds (vector-bounds (caddr sexp)))
		     (lowbnd (if bounds (min (cadr bounds) (caddr bounds)) 0))
		     (dvs (devectorize (caddr sexp) (+ lowbnd (cadr sexp)))))
		;;(if (not (zero? lowbnd)) (print 'logbit? sexp bounds lowbnd))
		;;(translate-exp bool? CONTLINE dvs)
		(translate-name dvs *verilog-output*)
		(cond (dvs (out CONTLINE #\[)
			   ;;(translate-exp bool? CONTLINE (+ lowbnd (cadr sexp)))
			   (out CONTLINE (+ lowbnd (cadr sexp)))
			   (out CONTLINE #\])))))
	     (else
	      (out CONTLINE 'LOGBIT?)
	      (infix-translate-list #f (+ 2 indent) (cdr sexp)))))
	   ((ARRAY-REF LOGAREF)
	    (translate-logaref indent (cadr sexp) (cddr sexp)))
	   ((NOT)
	    (out CONTLINE "!")
	    (optionally-parenthesize #t (+ 1 indent) (cadr sexp)))
	   ((OR)
	    (cond ((memq #t (cdr sexp)) (translate-exp #t indent #t))
		  (else (infix-translate-exp
			 #t " || " indent (remove #f (cdr sexp))))))
	   ((XOR)
	    (infix-translate-exp #t " ^ " indent (remove #f (cdr sexp))))
	   ((&&)
	    (infix-translate-exp #t " && " indent (cdr sexp)))
	   ((AND)
	    (cond ((memq #f (cdr sexp)) (translate-exp #t indent #f))
		  (else
		   (translate-and bool? indent (remove #t (cdr sexp))))))
	   ((IF)
	    (translate-if bool? indent sexp #t))
	   ((SYNCH:PRE)
	    (let ((expn (signal-table:get-next-fun *block* (cadr sexp))))
	      ;;(print 'synch:pre (cadr sexp) '*block* *block* '-> expn)
	      (if (not expn) (report "returned #f" sexp 'in *block*))
	      (translate-exp bool? indent
			     (or expn (symbol-append
				       *block* '- (cadr sexp) '?)))))
	   ((CASE CASEV)
	    (translate-case bool? indent (cadr sexp) (cddr sexp)))
	   ((CHECK=)
	    (translate-exp bool? indent (cadddr sexp)))
	   ((SYNCH:CHAT NUMBER-CHECK)
	    (out CONTLINE "1"))
	   ;;((SYNCH:PRE) (translate-exp bool? indent (macroexpand-1 sexp)))
	   (else (cond ((and (list? sexp) (= 2 (length sexp))
			     (apply signal-table:row-retrieve sexp))
			(translate-exp bool? indent (cadr sexp)))
		       ((pair? (car sexp)) ;computed function
			(out indent "(*(")
			(translate-exp #f (+ 3 indent) (car sexp))
			(out CONTLINE "))")
			(out (+ 2 indent))
			(infix-translate-list
			 #f (+ 2 indent) (cdr sexp)))
		       (else
			(out CONTLINE (car sexp))
			(infix-translate-list
			 #f (+ 2 indent) (cdr sexp)))))))))

(define (translate-if bool? indent sexp sense)
  (define test (cadr sexp))
  (define consequent (caddr sexp))
  (define alternate (cadddr sexp))
  (case (and (pair? test) (car test))
    ((NOT) (translate-if bool? indent
			 `(if ,(cadr test) ,consequent ,alternate)
			 (not sense)))
    ((DEFINED? PROVIDED?)
     (translate-ifdef test consequent alternate
		      (lambda (sexp)
			(if (negative? indent) (set! indent 16))
			(out indent)
			(translate-exp bool? (+ 2 indent) sexp))
		      sense))
    (else
     (if (not sense)
	 (let ((tmp consequent))
	   (set! consequent alternate)
	   (set! alternate tmp)))
     (optionally-parenthesize #t (+ 1 indent) test)
     (cond ((notany list? (cddr sexp))
	    (out CONTLINE " ? ")
	    (translate-exp bool? CONTLINE consequent)
	    (out CONTLINE " : ")
	    (translate-exp bool? CONTLINE alternate))
	   (else
	    (out (+ 4 indent) "? ")
	    (translate-exp bool? CONTLINE consequent)
	    (out (+ 4 indent) ": ")
	    (translate-exp bool? CONTLINE alternate))))))

(define (translate-ifdef test consequent alternate compile sense)
  (if (not sense)
      (let ((tmp consequent))
	(set! consequent alternate)
	(set! alternate tmp)))
  (out 0 (string-append "`ifdef "))
  (cond ((and (eq? 'PROVIDED? (car test))
	      (pair? (cadr test))
	      (eq? 'QUOTE (caadr test)))
	 (out CONTLINE
	      (string-upcase (symbol->string (cadadr test)))))
	((string? (cadr test))
	 (out CONTLINE (cadr test)))
	((symbol? (cadr test))
	 (out CONTLINE (cadr test)))
	(else
	 (report "argument not identifier" test)))
  (compile consequent)
  (out 0 "`else")
  (compile alternate)
  (out 0 "`endif")
  (out 0))

(define (translate-case bool? indent cv clauses)
  (cond ((null? clauses) (translate-exp bool? indent (if bool? #f 0)))
	((and bool? (or (not (assq 'ELSE clauses))
			(not (car (last-pair (assq 'ELSE clauses))))))
	 (translate-exp
	  bool? indent
	  (cons
	   'or (map (lambda (clause)
		      `(and
			,(cond ((eq? 'ELSE (car clause))
				(cond
				 ((car (last-pair clause))
				  (report "else not #f?" clause)))
				#t)
			       ((= 1 (length (car clause)))
				`(= ,(case-eval (caar clause) clause) ,cv))
			       (else
				`(or ,@(map (lambda (v)
					      `(= ,(case-eval v clause) ,cv))
					    (car clause)))))
			,@(if (and
			       (pair? (cadr clause))
			       (eq? 'and (caadr clause)))
			      (cdadr clause)
			      (list (cadr clause)))))
		    clauses))))
	(else
	 (let ((clause (car clauses)))
	   (translate-exp
	    bool? indent
	    (if (eq? 'ELSE (car clause))
		(car (last-pair clause))
		`(if ,(cond ((= 1 (length (car clause)))
			     `(= ,(case-eval (caar clause) clause) ,cv))
			    (else
			     `(or ,@(map (lambda (v)
					   `(= ,(case-eval v clause) ,cv))
					 (car clause)))))
		     ,(car (last-pair clause))
		     (case ,cv ,@(cdr clauses)))))))))

(define (ram-ref block array idxs)
  (cond ((ram-table:row-retrieve block array) =>
	 (lambda (ram-row)
	   (destructuring-bind
	       (block name output write-enable address data)
	       ram-row
	     (if (and (eq? (car idxs) address)
		      (null? (cdr idxs)))
		 (out CONTLINE output)
		 (report "Bad use of ram array:" array idxs)))))
	(else #f)))

(define (translate-logaref indent array idxs)
  ;;(print 'translate-logaref 'array array 'idxs idxs)
  (let* ((bounds (vector-bounds array))
	 (name (if bounds (car bounds) array)))
    ;;(print 'name name 'bounds bounds)
    (or (ram-ref *block* array idxs)
	(begin
	  (out CONTLINE name #\[)
	  (translate-exp #f indent (car idxs))
	  (out CONTLINE #\])))))

(define (optionally-parenthesize bool? indent x)
  (define bounds (and (symbol? x) (vector-bounds x)))
  (cond ((and bounds (not (<= 31 (abs (apply - (cdr bounds))))))
	 (out CONTLINE "{1'b0, ")
	 (translate-exp bool? (+ 1 indent) x)
	 (out CONTLINE "}"))
	((and (pair? x) (eqv? (car x) 'BIT-FIELD))
	 (out CONTLINE "{1'b0, ")
	 (translate-exp bool? (+ 1 indent) x)
	 (out CONTLINE "}"))
	((and (pair? x) (not (memq (car x)
				   '(NOT BIT-EXTRACT BIT-FIELD LOGBIT))))
	 (out CONTLINE #\()
	 (translate-exp bool? (+ 1 indent) x)
	 (out CONTLINE #\)))
	(else (translate-exp bool? indent x))))

(define (translate-and bool? indent exps)
  (case (length exps)
    ((0) (translate-exp #t indent #t))
    ((1) (translate-exp bool? indent (car exps)))
    (else (cond (bool? (infix-translate-exp #t " && " indent exps))
		(else (translate-exp
		       #f indent `(if (&& ,@(butlast exps 1))
				      ,(car (last-pair exps)) #f)))))))

(define (infix-translate-list bool? indent exps)
  (out CONTLINE #\()
  (cond ((not (null? exps))
	 (translate-exp bool? indent (car exps))
	 (set! exps (cdr exps))))
  (for-each (lambda (x)
	      (out CONTLINE #\, #\space)
	      (translate-exp bool? indent x))
	    exps)
  (out CONTLINE #\)))

(define (infix-translate-arith op indent exps)
  (define break #f)
  (cond ((not (null? exps))
	 (optionally-parenthesize #f indent (car exps))
	 (set! exps (cdr exps))))
  (set! break
	(if (some (lambda (x)
		    (and (pair? x)
			 (not (memq (car x)
				    '(NOT BIT-EXTRACT BIT-FIELD LOGBIT?)))))
		  exps)
	    indent
	    CONTLINE))
  (for-each (lambda (x)
	      (out break op)
	      (optionally-parenthesize
	       #f (+ (if (char? op) 1 (string-length op)) indent) x))
	    exps))

(define (infix-translate-exp bool? op indent exps)
  (define break #f)
  (cond ((not (null? exps))
	 (optionally-parenthesize bool? indent (car exps))
	 (set! exps (cdr exps))))
  (set! break
	(if (some (lambda (x)
		    (and (pair? x)
			 (not (memq (car x)
				    '(NOT BIT-EXTRACT BIT-FIELD LOGBIT?)))))
		  exps)
	    indent
	    CONTLINE))
  (for-each (lambda (x)
	      (out break op)
	      (optionally-parenthesize
	       bool? (+ (if (char? op) 1 (string-length op)) indent) x))
	    exps))

(define (translate-= indent num1 arg2)
  (cond ((and (number? arg2) (not (number? num1)))
	 (let ((tmp num1))
	   (set! num1 arg2)
	   (set! arg2 tmp))))
  (infix-translate-exp #f " == " indent (list num1 arg2)))

(define (translate-<> indent op num1 arg2)
  (infix-translate-exp #f (string-append " " (symbol->string op) " ")
		       indent (list num1 arg2)))

(define (translate-power-connections)
  (define (power? name)
    (or (memq name '(vcc gnd))
	(let* ((str (symbol->string name))
	       (n (string-length str)))
	  (and (> n 2)
	       (memv (string-ref str 0) '(#\+ #\-))
	       (char=? (string-ref str (- n 1)) #\v)))))
  (define (out-net r p-nets non-p-nets)
    (out 3 (if (memq 'gnd p-nets)
	       "pulldown "
	       "pullup ")
	 (symbol-append 'r r)
	 " (")
    (for-each (lambda (net)
		(out CONTLINE net ", "))
	      non-p-nets)
    (let loop ((p-nets p-nets))
      (cond ((null? (cdr p-nets))
	     (out CONTLINE (car p-nets) ");"))
	    (else
	     (out CONTLINE (car p-nets) ", ")
	     (loop (cdr p-nets))))))
  ;; Returns list of power or ground net names.
  (define (out-all-nets row)
    (apply
     (lambda (name index sheet u-pins r-pins)
       (for-each
	(lambda (r-net)
	  (let loop ((nets (cadr r-net))
		     (p-nets '())
		     (non-p-nets '()))
	    (cond ((null? nets)
		   (if (pair? p-nets)
		       (out-net (car r-net) p-nets non-p-nets)))
		  ((power? (car nets))
		   (loop (cdr nets)
			 (cons (car nets) p-nets)
			 non-p-nets))
		  (else
		   (loop (cdr nets)
			 p-nets
			 (cons (car nets) non-p-nets))))))
	(map resistor-nets:row-retrieve r-pins)))
     row))
  (define (print-net name bounds)
    ((netlist 'for-each-row-in-order)
     (lambda (row)
       (out-all-nets row))
     name #f))

  ((netlist-aggregate 'for-each-row-in-order)
   (lambda (row)
     (apply print-net row)))

  ((netlist-non-aggregate 'for-each-row-in-order)
   (lambda (row)
     (print-net (car row) #f))))

(define (translate-u-nets)
  (let ((nets '()))
    ((netlist 'for-each-row-in-order)
     (lambda (row)
       (apply (lambda (name index sheet u-pins r-pins)
		(if (not sheet)
		    (set! nets (cons (list name) nets))))
	      row)))
    (out-decl "wire" nets #f))
  (letrec ((netlist-inverse-for-each (netlist-inverse 'for-each-row-in-order))
	   (ptag-for-each
	    (lambda (ptag proc)
	      (netlist-inverse-for-each
	       (lambda (row) (apply proc row)) ptag))))
    ((ptag-table 'for-each-row-in-order)
     (lambda (row)
       (apply (lambda (ptag pld-type signature optcom)
		(let ((module #f)
		      (instance #f))
		  (sscanf (symbol->string ptag) "%[^-]-%[^-]" module instance)
		  (cond (signature
			 (set! instance module)
			 (set! module (symbol-append *design* '- module))))
		  (cond (instance
			 (out 3 module " " (or instance module) "(")
			 (let ((count 0))
			   (ptag-for-each
			    ptag
			    (lambda (ptag pin net-name index)
			      (if (> count 0) (out CONTLINE ", "))
			      (let ((col (if (= 4 (remainder count 5))
					     7
					     CONTLINE)))
				(set! count (+ 1 count))
				(cond (index
				       (out col #\. pin
					    "(" net-name "[" index "])"))
				      (else
				       (out col #\. pin "(" net-name ")")))))))
			 (out CONTLINE ");")
			 (out 0)))))
	      row)))))

;; (db:read-u-ptag "cz-umap.scm")
;; (db:read-net "cz")
(define (translate-netlist)
  (translate-u-nets)
  (translate-power-connections))

;;(print 'doing-if) (trace translate-assignment translate-top) (set! *qp-width* 333)
;;(trace-all "/home/jaffer/synch/scm2vrlg.scm") (set! *qp-width* 333) (untrace translate-name lookup registered-input? numbers-in-expression? ifable? infix-translate-exp merge-bounds get-bounds) ; translat!e-exp
